import scipy as _scipy                          
import pylab as _pylab
from scipy import interpolate as _interpolate
import spinmob as _s                      
import pylab_helper_standalones as _pylab_help
import time as _time
import spinmob_functions as _fun
import numpy as _numpy

class spline_single:
    # this class is a container for a single spline fit curve
    # initialize it with the results from a spline fit (that array thingy)

    # simply store the fit result on creation    
    def __init__(self, xdata, ydata, smoothing=5000, degree=5, presmoothing=0, plot=True, xlabel="x", ylabel="y", show_derivative=0, xmin="same", xmax="same", simple=0):
        """
        Create a spline object and fit xdata vs ydata with spline
        Set type="interpolate" to simply interpolate the raw data
        """

        # store this stuff for later use
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.xdata  = xdata
        self.ydata  = ydata
        self.simple = simple
        self.xmin = min(xdata)
        self.xmax = max(xdata)
        self._path = "(spline not saved)"
        self.smoothing    = smoothing
        self.degree       = degree
        self.presmoothing = presmoothing
        self.plot_flag    = plot
        self.message = ""
                
        
        # if we're not in simple mode, we have to do a spline fit, not just store the data
        if not simple:        
            # self.ydata might be smoothed
            self.ydata_smoothed = list(ydata)

            # if we're supposed to, presmooth the data
            if presmoothing: _fun.smooth_array(self.ydata_smoothed, presmoothing)

            print("presmoothing = ", str(presmoothing))
            print("smoothing = ",    str(smoothing))
            print("degree    = ",    str(degree))
        
            # do the fit
            self.pfit = _interpolate.splrep(xdata, self.ydata_smoothed, s=smoothing, k=degree)


        # now plot if we're supposed to
        if plot:
            fig  = _pylab.gcf()
            fig.clear()
            axes = fig.gca()
            axes.plot(xdata,ydata, "xr")
            axes.set_xlabel(xlabel)
            axes.set_ylabel(ylabel)
            if not simple:
                axes.plot(xdata,self.ydata_smoothed,"+g")
                self.plot(steps=len(xdata)*5, clear=False)
                _pylab_help.set_all_line_attributes("lw", 2)
                if show_derivative: self.plot(steps=len(xdata)*5, clear=False, derivative=show_derivative, yaxis='right')
                _pylab_help.set_all_line_attributes("lw", 1)
                _pylab_help.set_all_line_attributes("color", "r")

            _pylab_help.set_xrange(xmin, xmax)
            fig.canvas.Refresh()
        

        
    # this takes a single value or an array!
    def __call__(self,       x, derivative=0, smooth=0, simple='auto'):
        return self.evaluate(x, derivative,   smooth,   simple)
    def evaluate(self,       x, derivative=0, smooth=0, simple='auto'):
        """
        smooth=0      is how much to smooth the spline data
        simple='auto' is whether we should just use straight interpolation
                      you may want smooth > 0 for this, when derivative=1
        """
        if simple=='auto': simple = self.simple
        
        # make it into an array if it isn't one, and remember that we did
        is_array = True
        if not type(x) == type(_pylab.array([])):
            x = _pylab.array([x])              
            is_array = False

        if simple:
            # loop over all supplied x data, and come up with a y for each
            y = []
            for n in range(0, len(x)):
                # get a window of data around x
                if smooth:
                    [xtemp, ytemp, etemp] = _fun.trim_data(self.xdata, self.ydata, None, [x[n]-smooth, x[n]+smooth])
                else:
                    i1 = _fun.index_nearest(x[n], self.xdata)

                    # if the nearest data point is lower than x, use the next point to interpolate                    
                    if self.xdata[i1] <= x[n] or i1 <= 0: i2 = i1+1
                    else:                                 i2 = i1-1

                    # if we're at the max, extrapolate
                    if i2 >= len(self.xdata):
                        print(x[n], "is out of range. extrapolating")
                        i2 = i1-1

                    x1 = self.xdata[i1]
                    y1 = self.ydata[i1]
                    x2 = self.xdata[i2]
                    y2 = self.ydata[i2]
                    slope = (y2-y1)/(x2-x1)
                                        
                    xtemp = _numpy.array([x[n]])
                    ytemp = _numpy.array([y1 + (x[n]-x1)*slope])
                    
                # calculate the slope based on xtemp and ytemp (if smoothing)
                # or just use the raw slope if smoothing=0
                if derivative == 1:
                    if smooth:
                        y.append((_numpy.average(xtemp*ytemp)-_numpy.average(xtemp)*_numpy.average(ytemp)) /
                                 (_numpy.average(xtemp*xtemp)-_numpy.average(xtemp)**2))
                    else:
                        y.append(slope)
                        
                # otherwise just average (even with one element)
                elif derivative==0:
                    y.append(_numpy.average(ytemp))
    
            if is_array: return _numpy.array(y)
            else:        return y[0]
            
        
        if smooth:
            
            y = []
            for n in range(0, len(x)):
                # take 20 data points from x+/-smooth
                xlow = max(self.xmin,x[n]-smooth)
                xhi  = min(self.xmax,x[n]+smooth)
                xdata = _pylab.linspace(xlow, xhi, 20)
                ydata = _interpolate.splev(xdata, self.pfit, derivative)
                y.append(_numpy.average(ydata))

            if is_array: return _numpy.array(y)
            else:        return y[0]
            
        else:       
            return _interpolate.splev(x, self.pfit, derivative)

    
    def plot(self, derivative=0, xmin="auto", xmax="auto", steps=500, smooth=0, simple='auto', clear=True, yaxis='left'):
        if simple=='auto': simple = self.simple

        # get the min and max
        if xmin=="auto": xmin = self.xmin
        if xmax=="auto": xmax = self.xmax

        # get and clear the figure and axes
        f = _pylab.gcf()
        if clear and yaxis=='left': f.clf()

        # setup the right-hand axis
        if yaxis=='right':  a = _pylab.twinx()
        else:               a = _pylab.gca()

        # define a new simple function to plot, then plot it
        def f(x): return self.evaluate(x, derivative, smooth, simple)
        _pylab_help.plot_function(f, xmin, xmax, steps, clear, axes=a)

        # label it
        th = "th"
        if derivative == 1: th = "st"
        if derivative == 2: th = "nd"
        if derivative == 3: th = "rd"
        if derivative: self.ylabel = str(derivative)+th+" derivative of "+self.ylabel+" spline"
        a.set_xlabel(self.xlabel)
        a.set_ylabel(self.ylabel)
        a.figure.canvas.Refresh()
        

        


                


class spline_array:
    # this class holds an array of spline curves spline(x) at different values of some
    # "y" parameter, and will provide linear interpolation between them with evaluate()
    # evaluate_uber() is used when you generate y_splines.  I wouldn't use this yet.
    #
    def __init__(self, max_y_splines=100, simple=0):
        # create this class, then add spline curves to it    

        # the splines are stored in a dictionary with the key
        # as the parameter and the value is the spline_single
        self.x_splines = {}   # supplied by you
        self.y_splines = [{},{},{},{},{}] # generated by this class, index is x-derivative
        self.max_y_splines = max_y_splines # this sets the minimum x_parameter spacing of the y-splines
        self.xmin = None      # set the minimum and maximum values over which this is valid
        self.xmax = None
        self.ymin = None
        self.ymax = None
        self.xlabel = None
        self.ylabel = None
        self.zlabel = None
        self._path = "(spline array not saved)"
        self.simple=simple

    def add_x_spline(self, y_parameter, x_spline):
        # this function adds a spline to the dictionary
        self.x_splines[y_parameter] = x_spline

        # define or update range of validity
        if self.xmin == None: self.xmin = x_spline.xmin
        if self.xmax == None: self.xmax = x_spline.xmax
        if self.ymax == None: self.ymax = y_parameter
        if self.ymin == None: self.ymin = y_parameter

        if x_spline.xmax > self.xmax: self.xmax = x_spline.xmax
        if x_spline.xmin < self.xmin: self.xmin = x_spline.xmin
        if y_parameter > self.ymax: self.ymax = y_parameter
        if y_parameter < self.ymin: self.ymin = y_parameter

        # add in order to the master list of y_values
        self.y_values = list(self.x_splines.keys())
        self.y_values.sort()
        
    def remove_x_spline(self, y_parameter):
        # this function removes a spline from the dictionary
        try: self.x_splines.pop(y_parameter)
        except: print("Spline with parameter "+str(y_parameter)+" doesn't exist! Stop being a damn FOOL.")

    def __call__(self,       x, y, x_derivative=0, smooth=0, simple='auto'):
        return self.evaluate(x, y, x_derivative,   smooth,   simple)
    def evaluate(self,       x, y, x_derivative=0, smooth=0, simple='auto'):
        """
        this evaluates the 2-d spline by doing linear interpolation of the curves
        """

        if simple=='auto': simple = self.simple
        
        # find which values y is in between
        for n in range(0, len(self.y_values)-1):
            # if it's in between, interpolate!
            if self.y_values[n] <= y and self.y_values[n+1] >= y:
                y1 = self.y_values[n]
                y2 = self.y_values[n+1]
                z1 = self.x_splines[y1].evaluate(x, x_derivative, smooth, simple)
                z2 = self.x_splines[y2].evaluate(x, x_derivative, smooth, simple)
                return z1 + (y-y1)*(z2-z1)/(y2-y1)

        print("YARG! The y value "+str(y)+" is out of interpolation range!")            
        if y >= self.y_values[-1]: return self.x_splines[self.y_values[-1]].evaluate(x, x_derivative, smooth, simple)
        else                     : return self.x_splines[self.y_values[0]].evaluate(x, x_derivative, smooth, simple)

    
    def plot_fixed_x(self, x_values, x_derivative=0, steps=1000, smooth=0, simple='auto', ymin="auto", ymax="auto", format=True, clear=1):
        """
        plots the data at fixed x-value, so z vs x
        """
        if simple=='auto': simple=self.simple
        
        # get the min and max
        if ymin=="auto": ymin = self.ymin
        if ymax=="auto": ymax = self.ymax
        if clear: _pylab.gca().clear()

        if not type(x_values) in [type([]), type(_pylab.array([]))]: x_values = [x_values]

        for x in x_values:
    
            # define a new simple function to plot, then plot it
            def f(y): return self.evaluate(x, y, x_derivative, smooth, simple)
            _pylab_help.plot_function(f, ymin, ymax, steps, 0, False)

            # label it
            a = _pylab.gca()
            a.set_xlabel(self.ylabel)
            if x_derivative: a.set_ylabel(str(x_derivative)+" "+str(self.xlabel)+" derivative of "+self.zlabel)
            else:            a.set_ylabel(self.zlabel)
            a.set_title(self._path+"\nSpline array plot at fixed x = "+self.xlabel)
            a.get_lines()[-1].set_label("x ("+self.xlabel+") = "+str(x))

        if format: _s.format_figure()
        return a
    
    def plot_range_fixed_x(self, xmin="auto", xmax="auto", xsteps=21, ymin="auto", ymax="auto", ysteps=200, clear=True, x_derivative=0):
        if xmin=="auto": xmin=self.xmin
        if xmax=="auto": xmax=self.xmax
        
        self.plot_fixed_x(_pylab.linspace(xmin, xmax, xsteps), x_derivative, ysteps, ymin, ymax, False, clear)
        _s.format_figure()
        
    def plot_fixed_y(self, y_values, x_derivative=0, steps=1000, smooth=0, simple='auto', xmin="auto", xmax="auto", format=True, clear=1):
        """
        plots the data at a fixed y-value, so z vs y
        """
        if simple=='auto': simple=self.simple


        # get the min and max
        if xmin=="auto": xmin = self.xmin
        if xmax=="auto": xmax = self.xmax
        if clear: _pylab.gca().clear()

        if not type(y_values) in [type([]), type(_pylab.array([]))]: y_values = [y_values]

        for y in y_values:
            # define a new simple function to plot, then plot it
            def f(x): return self.evaluate(x, y, x_derivative, smooth, simple)
            _pylab_help.plot_function(f, xmin, xmax, steps, 0, True)

            # label it
            a = _pylab.gca()
            th = "th"
            if x_derivative == 1: th = "st"
            if x_derivative == 2: th = "nd"
            if x_derivative == 3: th = "rd"
            if x_derivative: a.set_ylabel(str(x_derivative)+th+" "+self.xlabel+" derivative of "+self.zlabel+" spline")
            else:            a.set_ylabel(self.zlabel)
            a.set_xlabel(self.xlabel)
            a.set_title(self._path+"\nSpline array plot at fixed y "+self.ylabel)
            a.get_lines()[-1].set_label("y ("+self.ylabel+") = "+str(y))
        if format: _s.format_figure()
        return a
    
    def plot_range_fixed_y(self, ymin="auto", ymax="auto", ysteps=21, xmin="auto", xmax="auto", xsteps=200, clear=True, x_derivative=0):
        if ymin=="auto": ymin=self.ymin
        if ymax=="auto": ymax=self.ymax
        
        self.plot_fixed_y(_pylab.linspace(ymin, ymax, ysteps), x_derivative, xsteps, xmin, xmax, False, clear)
        _s.format_figure()

    def generate_y_values(self):
        self.y_values = list(self.x_splines.keys())
        self.y_values.sort()


def copy_spline_array(a):
    """
    This returns an instance of a new spline_array with all the fixins, and the data from a.
    """

    b = spline_array()

    b.x_splines = a.x_splines
    b.y_splines = a.y_splines
    b.max_y_splines = a.max_y_splines
    b.xmin = a.xmin     
    b.xmax = a.xmax
    b.ymin = a.ymin
    b.ymax = a.ymax
    b.xlabel = a.xlabel
    b.ylabel = a.ylabel
    b.zlabel = a.zlabel
    b.simple = a.simple
    

    b.generate_y_values()    

    return b    

    
def load_spline_array(path=None, text="Give me a spline array to load, jerkface! No, YOU'RE the jerkface."):
    a = _s.load_object(path, text)

    b = copy_spline_array(a)
    b._path = a._path
    _s.save_object(b, b._path)

    return b    
            
def generate_spline_array(data, y_parameter="field", smoothing=5000, degree=5, presmoothing=0, coarsen=0, autosave=True, show_derivative=1, text="Give me some files to make a spline with.", simple=0):
    """

    Asks for a bunch of data files, plots and spline-fits the data (verifying at each step),
    and generates a spline_2d object instance filled with the results.

    """

    # create an instance of the spline_2d class
    s = spline_array()

    s.ylabel = y_parameter
    
    # have the user select a file
    paths = _s.DialogMultipleFiles('DIS AND DAT|*.dat', text=text, default_directory=data.directory)
    if paths == []: return

    # loop over each data file, fit it, plot it, ask if it's okay, and move on
    for n in range(0,len(paths)):
        print("spline "+str(n+1)+"/"+str(len(paths)+1))

        # fill up the xdata, ydata, and key            
        data.get_data(paths[n])

        print("y_parameter = "+str(data.constants[y_parameter]))

        # if this is just a simple interpolator, make a simple one
        if simple:
            x_spline = spline_single(data.xdata, data.ydata, plot=True, xlabel=data.xlabel, ylabel=data.ylabel, xmin="same", xmax="same", simple=1)

            a = input("ya: ")
            if a in ["quit", "q"]: return s
            if a in ["y", "yes", "\n"]: s.add_x_spline(data.constants[y_parameter], x_spline)

        # otherwise, we have to do the whole spline plot/fitting thingy
        else:
            data.plot(coarsen=coarsen)
            x_spline = splot("gca", smoothing, degree, presmoothing, interactive=True, show_derivative=show_derivative, simple=simple)
            
            if x_spline.message == "quit": return s
            if x_spline.message == "good": s.add_x_spline(data.constants[y_parameter], x_spline)

        s.xlabel = x_spline.xlabel
        s.zlabel = x_spline.ylabel

        # update the parameters as specified by the user
        presmoothing = x_spline.presmoothing
        smoothing    = x_spline.smoothing
        degree       = x_spline.degree

    print("Complete! "+str(len(paths))+" file and "+str(len(s.x_splines))+" successful splines.")

    s.simple = simple
    if autosave: _s.save_object(s)
    return s    



def splinteractive(xdata, ydata, smoothing=5000, degree=5, presmoothing=0, spline_class=spline_single, xlabel="x", ylabel="y", show_derivative=1, boost_factor=1.1, xmin="same", xmax="same"):
    """

    Give this function x and y data to spline fit and it will fit it, returning an instance of
    spline_single.  It will also ask you if it's okay and you can play with parameters.

    """    

    while True:        
        # fit it
        x_spline = spline_class(xdata, ydata, smoothing, degree, presmoothing, True, xlabel, ylabel, show_derivative, xmin=xmin, xmax=xmax)
        
        # ask if it's okay
        if x_spline.message == "": command = input("What now? ")
        else:                      command = x_spline.message
        
        try:
            float(command)
            command = "s="+command
        except:
            print("parsing...")

        # deal with simple commands
        if   command in ["y", "Y", "yes"]:
            x_spline.message = "good"
            smoothing = x_spline.smoothing
            return x_spline
        elif command in ["n", "N", "no"]:
            x_spline.message = "no"
            smoothing = x_spline.smoothing
            return x_spline
        elif command in ["q", "Q", "quit", "exit"]:
            x_spline.message = "quit"
            smoothing = x_spline.smoothing
            return x_spline
        elif command in ["l", "L", "line"]:
            x_spline.message = "line"
            smoothing = x_spline.smoothing
            x_spline.plot()
            input("(press <enter> when done looking)")
        elif command in ["d", "D", "derivative", "\n"]:
            x_spline.message = "derivative"
            smoothing = x_spline.smoothing
            x_spline.plot(derivative=1)
            input("(press <enter> when done looking)")
        elif command in ["dd", "DD"]:
            x_spline.message = "derivative 2"
            x_spline.plot(derivative=2)
            smoothing = x_spline.smoothing
            input("(press <enter> when done looking)")
        elif command in ["p", "P", "print", "printer"]:
            x_spline.message = "print"
            smoothing = x_spline.smoothing
            _s.printer()
        elif command == "[":
            x_spline.message = "unboost smoothing"
            smoothing = smoothing / boost_factor
        elif command == "]":
            x_spline.message = "boost smoothing"
            smoothing = smoothing * boost_factor

        # deal with parameter changes
        elif command.split("=")[0].strip() in ["s", "smoothing"]:
            try: smoothing = eval(command.split("=")[1].strip())
            except: print("Surely you can give me a better number than THAT piece of shit.")

        elif command.split("=")[0].strip() in ["d", "degree"]:
            try: degree = int(eval(command.split("=")[1].strip()))
            except: print("Nice try, ass.  Learn how to enter data.")

        elif command.split("=")[0].strip() in ["p", "pre", "presmoothing"]:
            try: presmoothing = int(eval(command.split("=")[1].strip()))
            except: print("Nice work.  Yeah, not really.")

        elif command.split("=")[0].strip() in ["b", "boost"]:
            try: boost_factor = float(eval(command.split("=")[1].strip()))
            except: print("Nice try smelly fart.")

        # idiot at the controls
        else:
            print("You need help.\n")
            print("------- COMMAND EXAMPLES -------\n")
            print("y        Yes, it looks good, move on.")
            print("n        No, and there's no hope.  Ignore and move on.")
            print("q        Quit.")
            print("s=1000   Set the smoothing to 10")
            print("d=3      Set the degree to 3")
            print("p=5      Set the presmoothing to 5")
            print("l        Show me the spline without data")
            print("d        Show me the derivative")
            print("]        Boost the smoothing")
            print("[        Reduce the smoothing")
            print("b=1.2    Set the boost factor to 1.2")

            smoothing = x_spline.smoothing


def splot(axes="gca", smoothing=5000, degree=5, presmoothing=0, plot=True, spline_class=spline_single, interactive=True, show_derivative=1):
    """
    
    gets the data from the plot and feeds it into splint
    returns an instance of spline_single

    axes="gca"                  which axes to get the data from.
    smoothing=5000              spline_single smoothing parameter
    presmoothing=0              spline_single data presmoothing factor (nearest neighbor)
    plot=True                   should we plot the result?
    spline_class=spline_single  which data class to use?
    interactive=False           should we spline fit interactively or just make a spline_single?
    
    """

    if   axes=="gca": axes = _pylab.gca()

    xlabel = axes.xaxis.label.get_text()
    ylabel = axes.yaxis.label.get_text()
    
    xdata = axes.get_lines()[0].get_xdata()
    ydata = axes.get_lines()[0].get_ydata()

    if interactive:
        return splinteractive(xdata, ydata, smoothing, degree, presmoothing, spline_class, xlabel, ylabel)
    else:
        return spline_class(xdata, ydata, smoothing, degree, presmoothing, plot, xlabel, ylabel)    






import os as _os
import spinmob as _s


def read_lines(path):
    f = open(path, 'rU')
    a = f.readlines()
    f.close()
    return(a)


class settings():
    path_user       = ''
    path_settings   = ''
    path_home       = ''
    path_temp       = ''
    path_colormaps  = ''
    _databox = _s.data.databox(delimiter='=')


    def __init__(self, name="spinmob"):
        """
        This class holds all the user-variables, paths etc...
        """

        # assemble the home and temp directory path for this environment
        self.path_user      = _os.path.expanduser('~')
        self.path_home      = _os.path.join(self.path_user, "."+name)
        self.path_settings  = _os.path.join(self.path_home, 'settings.txt')

        # see if this is the first time running (no home directory)
        if not _os.path.exists(self.path_home):
            print("\nFirst run:\n Creating "+self.path_home)
            _os.mkdir(self.path_home)

        if not _os.path.exists(self.path_settings):
            print(" Creating "+self.path_settings + "\n")
            f = open(self.path_settings, 'w')
            f.close()


        # now read in the prefs file
        self._databox.load_file(self.path_settings, header_only=True)


    def __call__   (self, key): return self.get(key)
    def __getitem__(self,key):  return self.get(key)
    def __setitem__(self,key,value): self.set(key, value)
    def __str__(self):
        s = ''
        for key in list(self._databox.hkeys):
            s = s + key + " = " + str(self._databox.h(key)) + '\n'
        return s

    def __repr__(self):
        s = '\nSPINMOB SETTINGS'
        keys = list(self._databox.hkeys)
        keys.sort()
        
        for key in keys:
            s = s + "\n  " + key + " = " + repr(self._databox.h(key))
        return s

    def keys(self):         return list(self._databox.hkeys)
    def has_key(self, key): return key in self._databox.hkeys

    def get(self, key):
        """
        Checks if the key exists and returns it. Returns None if it doesn't
        """
        if key in self._databox.hkeys: return self._databox.h(key)
        else:                          return None

    def _set_theme_figures(self, theme):    
        """
        Sets the matplotlib figure theme.
        
        Parameters
        ----------
        theme : str
            Can be 'classic' (for white) or 'dark_background' for dark theme
            (or any other theme that exists).
        """
        _s.pylab.style.use(theme)

    def set(self, key, value=None):
        """
        Sets the key-value pair and dumps to the preferences file.
        
        If value=None, pops / removes the setting.
        """
        if not key is None:            
            if not value is None: self._databox.h(**{key:value})
            else:                 self._databox.pop_header(key, ignore_error=True)

        # If figure theme is provided, update it.
        if key == 'dark_theme_figures':
            self._set_theme_figures('dark_background' if self['dark_theme_figures'] else 'classic')
            
            if not self['dark_theme_figures']: 
                _s.pylab.rcParams['figure.facecolor'] = 'white'
            
        # Other pylab settings
        if key in ['font_size']:
             _s.pylab.rcParams.update({
                 'font.size' : self['font_size']
                 })
        
        # Save the settings.
        self.save()


    def pop(self, *keys):
        """
        Pops the specified keys.
        """
        for key in keys: self.set(key)

    def clear(self):
        """
        Removes all settings.
        """
        self.pop(*self.keys())

    def make_dir(self, path="temp"):
        """
        Creates a directory of the specified path in the .spinmob directory.
        """
        full_path = _os.path.join(self.path_home, path)

        # only make it if it doesn't exist!
        if not _os.path.exists(full_path): _os.makedirs(full_path)

    def list_dir(self, path="temp"):
        """
        Returns a list of files in the specified path (directory), or an
        empty list if the directory doesn't exist.
        """
        full_path = _os.path.join(self.path_home, path)

        # only if the path exists!
        if _os.path.exists(full_path) and _os.path.isdir(full_path):
            return _os.listdir(full_path)
        else:
            return []

    def save(self):
        """
        Dumps the current prefs to the preferences.txt file
        """
        self._databox.save_file(self.path_settings, force_overwrite=True)





"""
Fit Multiple Data Sets
======================

Fitting multiple (simulated) Gaussian data sets simultaneously.

All minimizers require the residual array to be one-dimensional. Therefore, in
the ``objective`` we need to ```flatten``` the array before returning it.

TODO: this should be using the Model interface / built-in models!

"""
import spinmob as _s
import matplotlib.pyplot as plt
import numpy as _n

from lmfit import Parameters, minimize, report_fit

# Functions specified as strings
fstrings = ['a*x+b', 'a*x**2']
pnames   = ['a','b']

# List of functions from these strings.
fs = []
pstring  = ','.join(pnames)
for n in range(len(fstrings)): fs.append(eval('lambda x, '+pstring+': '+fstrings[n]))

# Function that returns the residuals in a 1D array
def residuals_concatenated(parameters, xdatas, ydatas):
    """
    Returns the 1D residuals for the specified parameters (p), xdata, and ydata.
    """
    # Get the list of parameter values
    ps = []
    for n in range(len(pnames)): ps.append(parameters[pnames[n]])
    
    # Calculate the residuals for each function using the supplied xdatas
    rs = []
    for n in range(len(xdatas)): rs.append(fs[n](_n.array(xdatas[n]), *ps)-ydatas[n])
        
    return _n.concatenate(rs)

# Data sets
xdatas = [[1,2,3,4], [1,2,3,4]]
ydatas = [[1,2,1,4], [1,3,2,2]]        

# Parameters
pfit = Parameters()
for pname in pnames: pfit.add(pname, value=1.0)

# Do the fit
results = minimize(residuals_concatenated, pfit, args=(xdatas, ydatas))
report_fit(results.params)




import os           as _os
import sys          as _sys
import matplotlib   as _mpl
import pylab


# Try to run the qt magic first
try:
    if not pylab.get_backend()[0:2] == 'Qt':
        get_ipython().run_line_magic('matplotlib', 'qt')
except: pass

import traceback    as _traceback
_sys.excepthook = _traceback.print_exception
_p = _traceback.print_last

# ignore warnings by default
import warnings as _warnings
_warnings.simplefilter("ignore")
_qtapp = None

# Import databox first
from . import _data      as data
from . import _functions as fun
from . import _thread    as thread
from . import _settings
settings = _settings.settings()
fun._settings  = settings
data._settings = settings

# Get the version
try: exec(fun.read_lines(_os.path.join(__path__[0],'setup.py'))[0])
except: __version__ = 'unknown'


def _warn(*args):
    if settings['ignore_warnings']: return

    args = list(args)
    print('\nWarning:', args.pop(0))
    for a in args: print('  ', a)
    print("  To disable warnings, set spinmob.settings['ignore_warnings']=True")


try:
    import pyqtgraph as _pyqtgraph
    _pyqtgraph_ok = float('.'.join(_pyqtgraph.__version__.split('.')[0:2])) > 0.1

    if not _pyqtgraph_ok: _warn('Spinmob requires pyqtgraph version 0.11 or higher. Previous versions are no longer supported.')

    _qtc = _pyqtgraph.Qt.QtCore
    _qt  = _pyqtgraph.Qt
    _qtw = _pyqtgraph.Qt.QtGui

    # make sure we have a valid qt application for dialogs etc...
    _qtapp = _qtc.QCoreApplication.instance()
    if not _qtapp: _qtapp = _qtw.QApplication([])

    # Set the dpi scaling
    _qtapp.setAttribute(_qtc.Qt.AA_EnableHighDpiScaling, True)

    # Standard Fusion light theme
    _qtapp.setStyle('Fusion')

    # Dark theme
    if settings['dark_theme_qt']: from . import _theme_fusion_dark


except:
    _warn("pyqtgraph version 0.11 or higher is required (use a conda or pip install).")

    _qtc = None
    _qt  = None
    _qtw = None
    _qtapp = None


# Try importing lmfit
try: import lmfit as _lm
except:
    _warn("spinmob.data.fitter() now requires lmfit, which is available via pip.")
    _lm = None


if _sys.version[0] == '2':
    _warn("Spinmob no longer supports Python 2.7 or below (which itself will not be maintained after Jan 1, 2020). USE AT YOUR OWN RISK!")


from . import _plot           as plot        ; plot._settings    = settings
from . import _dialogs        as dialogs     ; dialogs._settings = settings
from . import _pylab_tweaks   as tweaks      ; tweaks._settings  = settings

plot.tweaks._pylab_colormap._settings = settings

instaprint = tweaks.instaprint

# Default settings
_defaults = dict(
    dark_theme_qt         = False,
    dark_theme_figures    = False,
    font_size             = 12,
    font_size_legend      = 9)

for _k in _defaults:
    if not _k in settings.keys():
        settings[_k] = _defaults[_k]

# Touch the keys that affect visuals
for _k in [
    'font_size',
    'dark_theme_figures'
    ]:
    settings[_k] = settings[_k]








import numpy as _n          # This library adds all kinds of numerical functions
import spinmob.egg as egg   # this is our graphical window builder






#########################################
### Main window
#########################################

# Create a "sleg" window with the title 'Hello World!'
# This will not automatically show the window (see below)
window = egg.gui.Window('Hello World!', autosettings_path='example_other_widgets_window.cfg')

# make the second column the one that grows the most when resizing
window.set_column_stretch(1,100)

# top row (sub grid of misc controls defined below)
top_row = window.place_object(egg.gui.GridLayout(False), 0,0, column_span=2)




#########################################
### Buttons and file dialogs.
#########################################

# define a dummy functions to catch signals from buttons etc
def dummy_function(a): print("dummy function received argument:", a)
def open_files(a):     print(egg.dialogs.load_multiple(filters="*.txt"))

# add a buttons to the layout at grid position 0,0 (upper-left corner)
b0 = top_row.place_object(egg.gui.Button("File Dialog"), 0,0)

# connect the button's "clicked" signal to a function
window.connect(b0.signal_clicked, open_files)



#########################################
### Number box
#########################################

# make a number box, set its width
n = top_row.place_object(egg.gui.NumberBox(), 7,0, alignment=2)
n.set_width(50)

# connect this control's "changed" signal to the dummy function
# note: this is same as using window.connect() above.
window.connect(n.signal_changed, dummy_function)


#########################################
### Line Edit
#########################################
top_row.place_object(egg.gui.TextBox("TextBox Object."),8,0)


#########################################
### Toggle button
#########################################

# This button will toggle when pressed.
# The function it fires will receive its current state.
bt = top_row.place_object(egg.gui.Button("Checkable").set_checkable(), 5,0, alignment=1)

# yet another method to connect to a function
bt.signal_clicked.connect(dummy_function)


#########################################
### Label
#########################################

l = top_row.place_object(egg.gui.Label("Some text goes here."), 6,0, alignment=1)


#########################################
### pyqtgraph line plot and data taking
#########################################

# add a line plot in the second row
p1 = window.place_object(egg.pyqtgraph.PlotWidget(), 0,1)

# create a "box" to hold data and header information
data = egg.data.databox()

def acquire_fake_data(number_of_points=1000):
    """
    This function generates some fake data and returns two channels of data
    in the form

    time_array, [channel1, channel2]
    """
    # time array
    t = _n.linspace(0,10,number_of_points)

    return(t, [_n.cos(t)*(1.0+0.2*_n.random.random(number_of_points)),
               _n.sin(t      +0.5*_n.random.random(number_of_points))])

# function to acquire and plot some fake data
def acquire_and_plot(*a):

    # acquire fake data
    t, [y1,y2] = acquire_fake_data(1000)

    # plot fake data
    p1.clear()
    p1.plot(t, y1, pen=(1,2))
    p1.plot(t, y2, pen=(2,2))

    # store the data in the databox
    data['t']  = t
    data['y1'] = y1
    data['y2'] = y2

    # add header information
    data.insert_header('numberbox_value', n.get_value())

# create a button to acquire / plot data
b1 = top_row.place_object(egg.gui.Button("Acquire"), 1,0)

# connect the button's "clicked" signal to the "acquire data" function
window.connect(b1.signal_clicked, acquire_and_plot)




#########################################
### Saving data (can also load)
#########################################

# function to save the databox we created above
def save_data(a): data.save_file(filters="*.txt")

b2 = top_row.place_object(egg.gui.Button("Save Data"), 2,0)
window.connect(b2.signal_clicked, save_data)



#########################################
### pyqtgraph image plot
#########################################

# add a color plot to the second row spanning the right-most columns
p2 = window.place_object(egg.pyqtgraph.ImageView(), 1,1)

# function to plot an image
def randomize_image(*a): # note that *a will catch any number of arguments
                         # and store them all in a "tuple"
    # create a 2D random array
    z = _n.random.rand(100,100)

    # set the image data in the plot (see pyqtgraph documentation)
    p2.setImage(z)

# randomize it
#randomize_image()

# add a randomizer button above it
b3 = top_row.place_object(egg.gui.Button("Randomize"), 3,0)

# connect b3's click to re-randomizing the plot
window.connect(b3.signal_clicked, randomize_image)





#########################################
### Tab Area
#########################################

# add a tab area to the 2nd row, spanning all columns
tabs = window.place_object(egg.gui.TabArea(), 0,2, column_span=2)

# add two tabs
t1 = tabs.add_tab("Parameter Tree!")
t2 = tabs.add_tab("More Plots!")
t3 = tabs.add_tab("Plain Table!")
t4 = tabs.add_tab("Table Dictionary!")

# connect some tab signals
window.connect(tabs.signal_switched,            dummy_function)

# this signal is sent when the user clicks the red "x" on the tab
window.connect(tabs.signal_tab_close_requested, tabs.remove_tab)

# add a button and a plot to the first tab
t2.place_object(egg.pyqtgraph.PlotWidget(),0,0)
t2.place_object(egg.pyqtgraph.PlotWidget(),1,0)





#########################################
### Table
#########################################

# add the table object to the second tab
table = t3.place_object(egg.gui.Table(2,10))
window.connect(table.signal_cell_changed, dummy_function)

# a function that plays with the table
def do_something(a):
    table.set_value(2,11, 'test')
    table.set_value(1,1,  2.3)
    print(table.get_value(1,1))

# make a button
bt = t3.place_object(egg.gui.Button("Do something."),1,0)

# make the button do something
window.connect(bt.signal_clicked, do_something)


#########################################
# Table dictionary (souped up table)
# I use this for header information
#########################################

# create the dictionary
d = t4.place_object(egg.gui.TableDictionary(), 0,0).set_column_width(0,200).set_column_width(1,400).set_width(602)

# setting and getting values
d['test']         = 32
d['another test'] = [3,2,5]
d['final test']   = d['test']+400

# other ways to set / get. All values must be valid python code or it turns pink.
d(more_stuff="'strings need quotes'", comment_on_code='"this must be valid python code"')


#########################################
### Modified parameter tree.
#########################################

# create the parameter tree
tree = t1.place_object(egg.gui.TreeDictionary())

tree.add_parameter('LOL_WUT', 32.5, type='float')
tree.add_parameter('Some Category/parameter', '32')
tree.add_parameter('Some Category/parameter2', values=dict(a=32,b=45,c=17), value=45, type='list')
tree.add_parameter('LOL_WUT/test', 'lsdkjf')

tree.add_parameter('Numbers/limits(min=-7_max=0)', -3, type='int', limits=(-7,0), step=0.1)
tree.add_parameter('Numbers/Units Too!', 1e-6, type='float', siPrefix=True, suffix='V', step=1e-4)

x=tree.add_button('Some Category/test button')
x.signal_clicked.connect(dummy_function)

#########################################
### Show the window
#########################################

# show it
window.show(True) 

# If launching in spyder, it's good to call it without "True", so that you
# can interact with the thing on the command line as well as graphically.



# -*- coding: utf-8 -*-
"""
Module for testing _data.py
"""
import os      as _os # For loading fixtures
import numpy   as _n
import spinmob as _s

import unittest as _ut

a = b = c = d = x = y = f = None

class Test_fitter(_ut.TestCase):
    """
    Test class for fitter.
    """
    debug = False

    def setUp(self):
        # Path to the spinmob module
        self.data_path = _os.path.join(_os.path.dirname(_s.__file__), 'tests', 'fixtures')

        self.x1 = [0,1,2,3,4,5,6,7]
        self.y1 = [10,1,2,1,3,4,5,3]
        self.y2 = [2,1,2,4,5,2,1,5]
        self.ey = [0.3,0.5,0.7,0.9,1.1,1.3,1.5,1.7]

        self.x3 = [1,     2, 4.2,   4]
        self.y3 = [0.3, 0.5, 0.2, 0.7]
        self.plot_delay = 0.1

        return

    def test_basics_function_first(self):
        """
        Basic tests for a simple example smallish data set.
        """
        # Load a test file and fit it, making sure "f" is defined at each step.
        f = _s.data.fitter()
        f.__repr__()

        f.set_functions('a1 + a2*x + a3*x**2.', 'a1=-1., a2=0.04, a3=0.00006')
        f.__repr__()
        f.plot()

        f.set_data(self.x1, self.y1, 0.5)
        _s.pylab.ginput(timeout=self.plot_delay)
        f.__repr__()

        f.fit()
        _s.tweaks.set_figure_window_geometry(position=[0,0])
        _s.pylab.ginput(timeout=self.plot_delay)
        f.__repr__()

        # Check that the reduced chi^2 is roughly correct
        r = f.get_reduced_chi_squareds()
        self.assertIs(type(r), list)
        self.assertAlmostEqual(r[0], 29.2238, 2)

        # trim the data
        f.set(xmin=1.5, xmax=6.5)
        _s.pylab.ginput(timeout=self.plot_delay)
        f.__repr__()

        f.fit()
        _s.pylab.ginput(timeout=self.plot_delay)
        f.__repr__()

        # trim the data and test what happens when there are 0 DOF
        f.set(xmin=1.5, coarsen=2, plot_all_data=True, plot_guess_zoom=True)
        _s.pylab.ginput(timeout=self.plot_delay)
        f.__repr__()

        f.fit()
        _s.pylab.ginput(timeout=self.plot_delay)
        f.__repr__()

        # Change the first figure
        f(first_figure=1)

        # Change the function and starting plot, then refit
        f.set_functions('a*x+b', 'a,b')
        _s.tweaks.set_figure_window_geometry(position=[0,400])
        _s.pylab.ginput(timeout=self.plot_delay)
        f.__repr__()

        f.fit()
        _s.pylab.ginput(timeout=self.plot_delay)
        f.__repr__()

        # Coarsen and untrim

    def test_multi_data_sets(self):
        """
        Two-data-set-fit.

        Also includes
         - trim, zoom, etc
        """
        global f
        f = _s.data.fitter(first_figure=10)
        f.__repr__()

        # Set the data first
        f.set_data(self.x1, [self.y1,self.y2], self.ey)
        _s.tweaks.set_figure_window_geometry(10, position=[500,0])
        _s.tweaks.set_figure_window_geometry(11, position=[500,400])
        f.__repr__()

        # Set the functions
        f.set_functions(['a*x+b', 'a*cos(b*x)+c'], 'a=-1,b,c')
        f.__repr__()

        # Fit
        f.fit()
        f.__repr__()

        # Trim
        f.trim()
        f.__repr__()

        # Zoom
        f.zoom()
        f.__repr__()

        # Untrim
        f.untrim()
        f.__repr__()

        # Make sure untrim worked
        self.assertEqual(f['xmin'][0], None)

        # Fit
        f.fit()
        f.__repr__()

        # Two data sets with different x-datas, different lengths, and different bit depth
        f.set_data([self.x1, self.x1, self.x3], [self.y1,self.y2,self.y3], [self.ey,self.ey,45], dtype=_n.float32)
        f.__repr__()

        f.set_functions(['a*x+b', 'a*cos(b*x)+c', 'a*x**2'], 'a=-1,b,c')
        f.__repr__()

        # Fit it.
        f.fit()
        f.__repr__()

        # Set the wrong number of functions
        f.set_functions(['a*x+b', 'a*cos(b*x)+c'], 'a=-1,b,c')
        f.__repr__()

        f.set_functions(['a*x+b', 'a*cos(b*x)+c', 'a*x**2', 'a*x'], 'a=-1,b,c')
        f.__repr__()



    def test_get_processed_data(self):
        """
        Test self.get_processed_data().
        """

        f = _s.data.fitter(first_figure=7, autoplot=False)
        f.__repr__()

        # Set the data first
        f.set_data(self.x1, [self.y1,self.y2], self.ey)
        f.__repr__()

        # Massage conditions
        f(xmin=1.5, ymax=3, coarsen=2)
        f.__repr__()


        # Levels of process
        self.assertAlmostEqual(f.get_processed_data(                )[0][1][1], 6.5)
        self.assertAlmostEqual(f.get_processed_data(do_trim=False   )[0][1][3], 6.5)
        self.assertAlmostEqual(f.get_processed_data(do_coarsen=False)[2][0][3], 1.7)

    def test_dtype(self):
        """
        Ensures that the dtype is set.
        """

        f = _s.data.fitter(first_figure=7, autoplot=False)
        f.set_data(_n.array([1,2,3,4], dtype=_n.float16),
                   _n.array([1,2,3,4], dtype=_n.float64),
                   _n.array([1,2,3,4], dtype=_n.float32))
        self.assertEqual(f.get_processed_data()[0][0].dtype, _n.float16)

        f.set_data(_n.array([1,2,3,4], dtype=_n.float16),
                   _n.array([1,2,3,4], dtype=_n.float64),
                   _n.array([1,2,3,4], dtype=_n.float32),
                   dtype=_n.float32)
        self.assertEqual(f.get_processed_data()[1][0].dtype, _n.float32)


    def test_fix_free_and_function_globals(self):
        """
        Tests whether we can specify globals for the functions and do a fix()
        and free() call.
        """

        def my_fun(x,a,b): return a*x+b
        f = _s.data.fitter(autoplot=False)
        f.set_functions('stuff(x,a,b)', 'a,b', stuff=my_fun)
        f.fix('a')
        f.set_data([1,2,3],[1,2,1],0.1)
        f.fix(b=2)
        self.assertEqual(f['b'].value, 2)
        self.assertFalse(f['b'].vary)
        self.assertFalse(f['a'].vary)
        f.__repr__()
        f.fit()
        f.__repr__()
        f.free('a')
        f.__repr__()
        f.fit()
        f.__repr__()
        
    def test_background(self):
        f = _s.data.fitter(xmin=2, plot_guess_zoom=True, plot_all_data=True)
        f.set_data()
        f.set_functions('stuff*cos(x*other_stuff)+final_stuff', 'stuff, other_stuff, final_stuff', bg='final_stuff')
        f.fit()
    
    def test_autoscale_eydata(self):
        f = _s.data.fitter(autoplot=False).set_data(None, [[1,2,1,3],[1,2,1,4]]).set_functions(['a*x+b','a*cos(b*x)+c'], 'a,b,c').fit()
        f.autoscale_eydata().fit()
        self.assertAlmostEqual(f.results.redchi,0.9958597)

    def test_set_guess_to_fit_result(self):
        f = _s.data.fitter(autoplot=False).set_data(None, [[1,2,1,3],[1,2,1,4]]).set_functions(['a*x+b','a*cos(b*x)+c'], 'a,b,c').fit()
        f.set_guess_to_fit_result()
        self.assertEqual(f.p_fit['a'].value,f.p_in['a'].value)


if __name__ == "__main__":
    _ut.main(failfast=True)



import os                     as _os
import matplotlib             as _mpl
import pylab                  as _pylab
from functools import partial as _partial

try:    from . import _pylab_tweaks
except: import        _pylab_tweaks


import spinmob as _s
_qtw    = _s._qtw
_qt     = _s._qt
_qtcore = _s._qtc



# make sure we have an application
if __name__ == '__main__':
    _qtapp = _qtcore.QCoreApplication.instance()

    if not _qtapp:
        print("_pylab_colormap.py: Creating QApplication")
        _qtapp = _qtw.QApplication(_os.sys.argv)

    from . import _settings
    _settings = _settings.settings()


class colormap():

    # Each element of the list contains the position on the colorbar,
    # and the bottom/top colors
    _colorpoint_list = [[0.0, [1.0, 1.0, 1.0], [1.0, 1.0, 1.0]],
                        [0.5, [0.0, 0.0, 1.0], [0.0, 0.0, 1.0]],
                        [1.0, [1.0, 0.0, 0.0], [1.0, 0.0, 0.0]] ]

    # name of the colormap
    _name = "Last Used"

    # pylab image object
    _image=None

    def __init__(self, name="Last Used", image='auto'):
        """
        This object is responsible for loading and saving all the spinmob
        colormaps, and converting them to a cmap for pylab.
        """

        self.set_name(name)
        self.set_image(image)
        self.load_colormap()
        return

    def __repr__(self):
        s = "\n"+self.get_name()
        n = 0
        for x in self._colorpoint_list:
            s = s+"\n"+str(n)+": " + str(x[0])+" "+str(x[1])+" "+str(x[2])
            n += 1
        return s+"\n"

    def __getitem__(self, n): return self._colorpoint_list[n]

    def load_colormap(self, name=None):
        """
        Loads a colormap of the supplied name. None means used the internal
        name. (See self.get_name())
        """

        if name == None: name = self.get_name()
        if name == "" or not type(name)==str: return "Error: Bad name."

        # assemble the path to the colormap
        path = _os.path.join(_settings.path_home, "colormaps", name+".cmap")

        # make sure the file exists
        if not _os.path.exists(path):
            print("load_colormap(): Colormap '"+name+"' does not exist. Creating.")
            self.save_colormap(name)
            return

        # open the file and get the lines
        f = open(path, 'r')
        x = f.read()
        f.close()

        try:
            self._colorpoint_list = eval(x)
        except:
            print("Invalid colormap. Overwriting.")
            self.save_colormap()

        # update the image
        self.update_image()

        return self

    def save_colormap(self, name=None):
        """
        Saves the colormap with the specified name. None means use internal
        name. (See get_name())
        """
        if name == None: name = self.get_name()
        if name == "" or not type(name)==str: return "Error: invalid name."

        # get the colormaps directory
        colormaps = _os.path.join(_settings.path_home, 'colormaps')

        # make sure we have the colormaps directory
        _settings.make_dir(colormaps)

        # assemble the path to the colormap
        path = _os.path.join(_settings.path_home, 'colormaps', name+".cmap")

        # open the file and overwrite
        f = open(path, 'w')
        f.write(str(self._colorpoint_list))
        f.close()

        return self

    def delete_colormap(self, name=None):
        """
        Deletes the colormap with the specified name. None means use the internal
        name (see get_name())
        """
        if name == None: name = self.get_name()
        if name == "" or not type(name)==str: return "Error: invalid name."

        # assemble the path to the colormap
        path = _os.path.join(_settings.path_home, 'colormaps', name+".cmap")
        _os.unlink(path)

        return self

    def set_name(self, name="My Colormap"):
        """
        Sets the name.

        Make sure the name is something your OS could name a file.
        """
        if not type(name)==str:
            print("set_name(): Name must be a string.")
            return
        self._name = name
        return self

    def get_name(self):
        """
        Returns the name of the current colormap.
        """
        return self._name

    def set_image(self, image='auto'):
        """
        Set which pylab image to tweak.
        """
        if image=="auto": image = _pylab.gca().images[0]
        self._image=image
        self.update_image()

    def update_image(self):
        """
        Set's the image's cmap.
        """
        if self._image:
            self._image.set_cmap(self.get_cmap())
            _pylab.draw()

    def pop_colorpoint(self, n=0):
        """
        Removes and returns the specified colorpoint. Will always leave two behind.
        """

        # make sure we have more than 2; otherwise don't pop it, just return it
        if len(self._colorpoint_list) > 2:

            # do the popping
            x = self._colorpoint_list.pop(n)

            # make sure the endpoints are 0 and 1
            self._colorpoint_list[0][0]  = 0.0
            self._colorpoint_list[-1][0] = 1.0

            # update the image
            self.update_image()

            return x

        # otherwise just return the indexed item
        else: return self[n]

    def insert_colorpoint(self, position=0.5, color1=[1.0,1.0,0.0], color2=[1.0,1.0,0.0]):
        """
        Inserts the specified color into the list.
        """
        L = self._colorpoint_list

        # if position = 0 or 1, push the end points inward
        if   position <= 0.0:
            L.insert(0,[0.0,color1,color2])

        elif position >= 1.0:
            L.append([1.0,color1,color2])

        # otherwise, find the position where it belongs
        else:

            # loop over all the points
            for n in range(len(self._colorpoint_list)):
                # check if it's less than the next one
                if position <= L[n+1][0]:
                    # found the place to insert it
                    L.insert(n+1,[position,color1,color2])
                    break

        # update the image with the new cmap
        self.update_image()

        return self

    def modify_colorpoint(self, n, position=0.5, color1=[1.0,1.0,1.0], color2=[1.0,1.0,1.0]):
        """
        Changes the values of an existing colorpoint, then updates the colormap.
        """
        if    n==0.0                         : position = 0.0
        elif  n==len(self._colorpoint_list)-1: position = 1.0
        else: position = max(self._colorpoint_list[n-1][0], position)

        self._colorpoint_list[n] = [position, color1, color2]
        self.update_image()
        self.save_colormap("Last Used")

    def get_cmap(self):
        """
        Generates a pylab cmap object from the colorpoint data.
        """

        # now generate the colormap from the ordered list
        r = []
        g = []
        b = []
        for p in self._colorpoint_list:
            r.append((p[0], p[1][0]*1.0, p[2][0]*1.0))
            g.append((p[0], p[1][1]*1.0, p[2][1]*1.0))
            b.append((p[0], p[1][2]*1.0, p[2][2]*1.0))

        # store the formatted dictionary
        c = {'red':r, 'green':g, 'blue':b}

        # now set the dang thing
        return _mpl.colors.LinearSegmentedColormap('custom', c)






class colormap_interface(colormap):

    def __init__(self, name="Last Used", image="auto"):
        """
        This is the graphical interface for interacting with a
        pylab image.
        """
        colormap.__init__(self,name,image)

        # create the main window
        self._window = _qtw.QMainWindow()
        self._window.setWindowTitle("Colormap")

        # main widget inside window
        self._central_widget = _qtw.QWidget()
        self._window.setCentralWidget(self._central_widget)

        # add all the controls
        self._build_gui()

        # disable the save (just loaded)
        self._button_save.setEnabled(False)

        # set the location
        pos, size = _pylab_tweaks.get_figure_window_geometry()
        self._window.move(pos[0]+size[0],pos[1])

        self.show()



    def _build_gui(self):
        """
        Removes all existing sliders and rebuilds them based on the colormap.
        """
        # remove all widgets (should destroy all children too)
        self._central_widget.deleteLater()

        # remove all references to other controls
        self._sliders               = []
        self._buttons_top_color     = []
        self._buttons_bottom_color  = []
        self._checkboxes            = []
        self._buttons_plus          = []
        self._buttons_minus         = []
        self._color_dialogs_top     = []
        self._color_dialogs_bottom  = []

        # create the new central widget
        self._central_widget = _qtw.QWidget()
        self._window.setCentralWidget(self._central_widget)

        # layout for main widget
        self._layout = _qtw.QGridLayout(self._central_widget)
        self._central_widget.setLayout(self._layout)

        # add the list of cmaps
        self._combobox_cmaps = _qtw.QComboBox(self._central_widget)
        self._combobox_cmaps.setEditable(True)
        self._load_cmap_list()

        # add the save and delete buttons
        self._button_save   = _qtw.QPushButton("Save",   self._central_widget)
        self._button_delete = _qtw.QPushButton("Delete", self._central_widget)
        self._button_save.setFixedWidth(70)
        self._button_delete.setFixedWidth(70)

        # layouts
        self._layout.addWidget(self._combobox_cmaps, 1,1, 1,3, _qtcore.Qt.Alignment(0))
        self._layout.addWidget(self._button_save,    1,5, 1,1, _qtcore.Qt.Alignment(1))
        self._layout.addWidget(self._button_delete,  1,6, 1,2, _qtcore.Qt.Alignment(1))

        # actions
        self._combobox_cmaps.currentIndexChanged.connect(self._signal_load)
        self._button_save  .clicked.connect(self._button_save_clicked)
        self._button_delete.clicked.connect(self._button_delete_clicked)

        # ensmallen the window
        self._window.resize(10,10)

        # now create a control set for each color point
        for n in range(len(self._colorpoint_list)):

            c1 = self._colorpoint_list[n][1]
            c2 = self._colorpoint_list[n][2]

            # create a top-color button
            self._buttons_top_color.append(_qtw.QPushButton(self._central_widget))
            self._buttons_top_color[-1].setStyleSheet("background-color: rgb("+str(int(c2[0]*255))+","+str(int(c2[1]*255))+","+str(int(c2[2]*255))+"); border-radius: 3px;")

            # create a bottom-color button
            self._buttons_bottom_color.append(_qtw.QPushButton(self._central_widget))
            self._buttons_bottom_color[-1].setStyleSheet("background-color: rgb("+str(int(c1[0]*255))+","+str(int(c1[1]*255))+","+str(int(c1[2]*255))+"); border-radius: 3px;")

            # create color dialogs
            self._color_dialogs_top.append(_qtw.QColorDialog(self._central_widget))
            self._color_dialogs_top[-1].setCurrentColor(self._buttons_top_color[-1].palette().color(1))

            self._color_dialogs_bottom.append(_qtw.QColorDialog(self._central_widget))
            self._color_dialogs_bottom[-1].setCurrentColor(self._buttons_top_color[-1].palette().color(1))

            # create link checkboxes
            self._checkboxes.append(_qtw.QCheckBox(self._central_widget))
            self._checkboxes[-1].setChecked(c1==c2)

            # create a slider
            self._sliders.append(_qtw.QSlider(self._central_widget))
            self._sliders[-1].setOrientation(_qtcore.Qt.Horizontal)
            self._sliders[-1].setMaximum(1000)
            self._sliders[-1].setValue(int(self._colorpoint_list[n][0]*1000))
            self._sliders[-1].setFixedWidth(250)

            # create + and - buttons
            self._buttons_plus.append(_qtw.QPushButton(self._central_widget))
            self._buttons_plus[-1].setText("+")
            self._buttons_plus[-1].setFixedWidth(25)

            self._buttons_minus.append(_qtw.QPushButton(self._central_widget))
            self._buttons_minus[-1].setText("-")
            self._buttons_minus[-1].setFixedWidth(25)

            # layout
            self._layout.addWidget(self._buttons_bottom_color[-1], n+3,1,      _qtcore.Qt.AlignCenter)
            self._layout.addWidget(self._checkboxes[-1],           n+3,2, 1,1, _qtcore.Qt.AlignCenter)
            self._layout.addWidget(self._buttons_top_color[-1],    n+3,3,      _qtcore.Qt.AlignCenter)
            self._layout.addWidget(self._sliders[-1],              n+3,4, 1,2, _qtcore.Qt.AlignCenter)
            self._layout.setColumnStretch(5,100)
            self._layout.addWidget(self._buttons_minus[-1],        n+3,7,      _qtcore.Qt.AlignCenter)
            self._layout.addWidget(self._buttons_plus[-1],         n+3,6,      _qtcore.Qt.AlignCenter)

            # connect the buttons and slider actions to the calls
            self._buttons_bottom_color[-1]            .clicked.connect(_partial(self._color_button_clicked, n, 0))
            self._buttons_top_color[-1]               .clicked.connect(_partial(self._color_button_clicked, n, 1))
            self._color_dialogs_bottom[-1].currentColorChanged.connect(_partial(self._color_dialog_changed, n, 0))
            self._color_dialogs_top[-1]   .currentColorChanged.connect(_partial(self._color_dialog_changed, n, 1))

            self._buttons_plus[-1]        .clicked.connect(_partial(self._button_plus_clicked,  n))
            self._buttons_minus[-1]       .clicked.connect(_partial(self._button_minus_clicked, n))
            self._sliders[-1]        .valueChanged.connect(_partial(self._slider_changed,   n))



        # disable the appropriate sliders
        self._sliders[0] .setDisabled(True)
        self._sliders[-1].setDisabled(True)


    def _signal_load(self):
        """
        Load the selected cmap.
        """

        # set our name
        self.set_name(str(self._combobox_cmaps.currentText()))

        # load the colormap
        self.load_colormap()

        # rebuild the interface
        self._build_gui()

        self._button_save.setEnabled(False)


    def _button_save_clicked(self):
        """
        Save the selected cmap.
        """
        self.set_name(str(self._combobox_cmaps.currentText()))
        self.save_colormap()
        self._button_save.setEnabled(False)
        self._load_cmap_list()

    def _button_delete_clicked(self):
        """
        Save the selected cmap.
        """
        name = str(self._combobox_cmaps.currentText())
        self.delete_colormap(name)
        self._combobox_cmaps.setEditText("")
        self._load_cmap_list()


    def _color_dialog_changed(self, n, top, c):
        """
        Updates the color of the slider.
        """
        self._button_save.setEnabled(True)

        cp = self._colorpoint_list[n]

        # if they're linked, set both
        if self._checkboxes[n].isChecked():
            self.modify_colorpoint(n, cp[0], [c.red()/255.0, c.green()/255.0, c.blue()/255.0],
                                             [c.red()/255.0, c.green()/255.0, c.blue()/255.0])
            self._buttons_top_color   [n].setStyleSheet("background-color: rgb("+str(c.red())+","+str(c.green())+","+str(c.green())+"); border-radius: 3px;")
            self._buttons_bottom_color[n].setStyleSheet("background-color: rgb("+str(c.red())+","+str(c.green())+","+str(c.green())+"); border-radius: 3px;")


        elif top:
            self.modify_colorpoint(n, cp[0], cp[1], [c.red()/255.0, c.green()/255.0, c.blue()/255.0])
            self._buttons_top_color   [n].setStyleSheet("background-color: rgb("+str(c.red())+","+str(c.green())+","+str(c.green())+"); border-radius: 3px;")

        else:
            self.modify_colorpoint(n, cp[0], [c.red()/255.0, c.green()/255.0, c.blue()/255.0], cp[2])
            self._buttons_bottom_color[n].setStyleSheet("background-color: rgb("+str(c.red())+","+str(c.green())+","+str(c.green())+"); border-radius: 3px;")



    def _button_plus_clicked(self, n):
        """
        Create a new colorpoint.
        """
        self._button_save.setEnabled(True)

        self.insert_colorpoint(self._colorpoint_list[n][0],
                               self._colorpoint_list[n][1],
                               self._colorpoint_list[n][2])
        self._build_gui()

    def _button_minus_clicked(self, n):
        """
        Remove a new colorpoint.
        """
        self._button_save.setEnabled(True)

        self.pop_colorpoint(n)
        self._build_gui()

    def _slider_changed(self, n):
        """
        updates the colormap / plot
        """
        self._button_save.setEnabled(True)

        self.modify_colorpoint(n, self._sliders[n].value()*0.001, self._colorpoint_list[n][1], self._colorpoint_list[n][2])


    def _color_button_clicked(self, n,top):
        """
        Opens the dialog.
        """
        self._button_save.setEnabled(True)

        if top: self._color_dialogs_top[n].open()
        else:   self._color_dialogs_bottom[n].open()


    def _load_cmap_list(self):
        """
        Searches the colormaps directory for all files, populates the list.
        """
        # store the current name
        name = self.get_name()

        # clear the list
        self._combobox_cmaps.blockSignals(True)
        self._combobox_cmaps.clear()

        # list the existing contents
        paths = _settings.list_dir('colormaps')

        # loop over the paths and add the names to the list
        for path in paths:
            self._combobox_cmaps.addItem(_os.path.splitext(path)[0])

        # try to select the current name
        self._combobox_cmaps.setCurrentIndex(self._combobox_cmaps.findText(name))
        self._combobox_cmaps.blockSignals(False)

    def close(self):
        """
        Closes the window.
        """
        self._window.close()
        #_qt.QtWidgets.qApp.processEvents()

    def show(self):
        """
        Shows the window.
        """
        self._window.show()
        #_qt.QtWidgets.qApp.processEvents()


######################
## Example Code
######################







from spinmob import egg

# create the window and two tab areas
w     = egg.gui.Window('Hey Mammal II', autosettings_path='w')
tabs1 = w.place_object(egg.gui.TabArea(autosettings_path='tabs1'))
tabs2 = w.place_object(egg.gui.TabArea(autosettings_path='tabs2'), alignment=0)

# add some tabs
t_settings  = tabs1.add_tab("Settings")
t_raw       = tabs2.add_tab("Raw")
t_processed = tabs2.add_tab("Processed")






###########################
# TREE DICTIONARY
###########################
settings = t_settings.place_object(egg.gui.TreeDictionary('settings'))

# add some settings
settings.add_parameter('DAQ/Rate',      1000,      type='float', step=1000, suffix='Hz', siPrefix=True, limits=[1,1e5])
settings.add_parameter('DAQ/Pants',        0,      type='list',  values=['required','optional'])
settings.add_parameter('Other/Function', 'cos(t)', type='str')



#############################
# Databox Plotter
#############################
d_raw = t_raw.place_object(egg.gui.DataboxPlot('*.raw', 'd_raw'), alignment=0)

# Add some ROIs
r1 = egg.pyqtgraph.LinearRegionItem([1.5,3.5])
r2 = egg.pyqtgraph.InfiniteLine(4.2, movable=True)
r3 = egg.pyqtgraph.InfiniteLine(4.2, movable=True)
d_raw.ROIs = [[r1,r2], r3]

# Add some data / header info
d_raw['t']  = [1,2,3,4,5]
d_raw['V1'] = [1,2,1,2,1]
d_raw['V2'] = [2,1,2,1,2]
d_raw.h(Time="Pants O'Clock", Amp_Gain="32")

# Plot it
d_raw.plot()

# Overload the post-data-load function to send header info to s
def after_load_file(d): settings.update(d.headers, ignore_errors=True)
d_raw.after_load_file = after_load_file




# show it!
w.show(True)



import numpy as _n
from spinmob import egg

# Create the window
w = egg.gui.Window("Hey, mammal.")

# Move it to the upper left corner (plus 17 pixels)
w.set_position([17,0])

# Add a button
b = egg.gui.Button("Fire!")
w.place_object(b)

# Modify the button using the "daisy chain"
b.set_text("Fire Once!")

# Define a basic function
def f(*a): print("Oooooh, received", a)

# Connect the button to f
w.connect(b.signal_clicked, f)

# Create a databox with load / save buttons and place it in the window
d = egg.gui.DataboxSaveLoad(autosettings_path='d.cfg')
w.place_object(d)

# Add some basic data, and enable the save button
d['x'] = [1,2,3]
d['y'] = [1,2,1]
d.enable_save()

# New auto-row
w.new_autorow()

# (Mis-)Place the graph
p = w.place_object(egg.pyqtgraph.PlotWidget())

# Re-place it properly
w.place_object(p, 0,1, column_span=2, alignment=0)

# Set the column stretch
w.set_column_stretch(1, 10)

# Add a curve to the graph
c = egg.pyqtgraph.PlotCurveItem()
p.addItem(c)

# Define a function to get fake data
def get_fake_data(*a):
    """
    Called whenever someone presses the "fire" button.
    """
    # add columns of data to the databox
    d['x'] = _n.linspace(0,10,100)
    d['y'] = _n.cos(d['x']) + 0.1*_n.random.rand(100)
    
    # update the curve    
    c.setData(d['x'], d['y'])
 
# Connect the button to this function
b.signal_clicked.connect(get_fake_data)

# Create a custom post_load function
def my_post_load(): 
    """
    Called after someone loads data using the "load" button.
    """    
    # update the curve with the loaded data    
    c.setData(d['x'], d['y'])

# Overwrite the existing (blank) post-load function
d.post_load = my_post_load

# Show the window (we block the command line so it can be run by double-clicking)
w.show(block_command_line=True)


from . import _plotting_mess

data        = _plotting_mess.magphase_data
databoxes   = _plotting_mess.magphase_databoxes
files       = _plotting_mess.magphase_files
function    = _plotting_mess.magphase_function

_settings = None


from . import _pylab_tweaks as tweaks; reload(tweaks)
from . import _plotting_mess;          reload(_plotting_mess)






databoxes           = _plotting_mess.databoxes_xy
files               = _plotting_mess.files_xy
data                = _plotting_mess.xy
data_magphase       = _plotting_mess.mag_phase
data_realimag       = _plotting_mess.real_imag
surface_data        = _plotting_mess.xyz
function            = _plotting_mess.function_1D
function_parametric = _plotting_mess.function_parametric
surface_function    = _plotting_mess.function_2D
style               = _plotting_mess.plot_style_cycle


##########################################
# Histogram DataboxProcessor Plot Script #
##########################################

# List of datasets you'd like to plot, e.g. 'V1' (not 'V1.values', 'V1.counts', etc)
plot_columns = []

# Leaving plot_columns = [] will plot them all.

##############################
# No need to edit below this #
##############################
x = None
y = []

# If none are supplied, get them all!
if plot_columns==None or len(plot_columns) == 0:
  for n in range(len(d.ckeys)):
    k = d.ckeys[n]
    if k[len(k)-7:] == '.values': 
      plot_columns.append(k[0:len(k)-7])

# Now generate the data to plot and axes labels
x  = []
y  = []
ey = []
xlabels = []
ylabels = []
for k in plot_columns:
  
  # Data to plot
  x.append(d[k+'.values'])
  y.append(d[k+'.counts'])
  if k+'.counts.std_mean' in d.ckeys:
    ey.append(d[k+'.counts.std_mean'])

  # Axes labels
  xlabels.append(k)
  ylabels.append(k+'.counts')

# If we have no ey's, kill em so the script doesn't crash.
if len(ey) == 0: ey = None



import numpy as _n
import time  as _t
import spinmob.egg as egg



##### GUI DESIGN

# create the main window
w = egg.gui.Window(autosettings_path="example_sweeper_w")

# add the "go" button
b_sweep  = w.place_object(egg.gui.Button("Sweep!",         checkable=True))
b_select = w.place_object(egg.gui.Button("Use Blue Range", checkable=True))
l_x      = w.place_object(egg.gui.Label("x:"), alignment=2)
n_x      = w.place_object(egg.gui.NumberBox(int=False))

# move to the second row and add a TreeDictionary for our "settings"
w.new_autorow()
settings = w.place_object(egg.gui.TreeDictionary('example_sweeper_settings'), column_span=4, alignment=0)
settings.add_parameter('sweep/x_start', 0.0)
settings.add_parameter('sweep/x_stop',  0.004)
settings.add_parameter('sweep/x_steps', 200)

# add a tabbed interface for the plotting area, spanning the first and second rows
tabs = w.place_object(egg.gui.TabArea('example_sweeper_tabs'), 10,0, row_span=2, alignment=0)
w.set_column_stretch(10)

# add a tab for some plots
t_raw      = tabs.add_tab("Raw Plots")
t_analysis = tabs.add_tab("Analysis")

# add a databox plotter object to the tab
d_sweep = t_raw.place_object(egg.gui.DataboxPlot(autosettings_path="example_sweeper_d_sweep"), alignment=0)
d_sweep['x']     = []
d_sweep['mag']   = []
d_sweep['phase'] = []

# add a "region of interest" (ROI) for selecting the sweep range
roi_sweep = egg.pyqtgraph.LinearRegionItem([settings['sweep/x_start'], settings['sweep/x_stop']])
d_sweep.ROIs = [roi_sweep]

# show the blank plots
d_sweep.plot()


##### MAIN FUNCTIONALITY

# define a function to set some parameter on an external instrument
def set_x(x):
    """
    Pretends to set some instrument to a value "x" somehow.
    This is where your code should go.
    """

    # for now just update the gui
    n_x.set_value(x)

# define a function to create fake data.
# This is where you might put code for talking to a piece of equipment
def get_data():
    """
    Currently pretends to talk to an instrument and get back the magnitud
    and phase of the measurement.
    """

    # pretend we're measuring a noisy resonance at zero
    y = 1.0 / (1.0 + 1j*(n_x.get_value()-0.002)*1000) + _n.random.rand()*0.1

    # and that it takes time to do so
    _t.sleep(0.1)

    # return mag phase
    return abs(y), _n.angle(y, True)


# define a function to be called whenever the acquire button is pressed
def acquire_button_clicked(*a):

    # don't start another loop if the button is unchecked!
    if not b_sweep.is_checked(): return

    # disable the settings during the sweep
    settings.disable()

    # figure out the sweep range from the graph if we're supposed to
    if b_select.is_checked():
        settings['sweep/x_start'], settings['sweep/x_stop'] = roi_sweep.getRegion()

    # otherwise update the roi_sweep to match the sweep
    else: roi_sweep.setRegion((settings['sweep/x_start'],
                               settings['sweep/x_stop']))

    # clear the data and create new columns
    d_sweep.clear()
    d_sweep['x']     = []
    d_sweep['mag']   = []
    d_sweep['phase'] = []

    # dump the settings to the databox header
    settings.send_to_databox_header(d_sweep)

    # start the loop and keep looping until someone
    # unchecks the acquisition button or we max out the iterations
    # setting iterations = 0 will loop infinitely
    for x in _n.linspace(settings['sweep/x_start'],
                         settings['sweep/x_stop'],
                         settings['sweep/x_steps']):

        # make sure the button is still pressed
        if not b_sweep.is_checked(): break

        # set the current value
        set_x(x)

        # get the data
        m, p = get_data()

        # store the data
        d_sweep.append_row([x,m,p])

        # plot the data
        d_sweep.plot()

        # process other window events so the GUI doesn't freeze!
        w.process_events()

    # in case the button is still checked
    b_sweep.set_checked(False)

    # re-enable the settings
    settings.enable()

# connect the button
b_sweep.signal_clicked.connect(acquire_button_clicked)




######### OTHER GUI STUFF


# overwrite the post_load function so it plots and sets up the settings
def d_sweep_after_load():

    # dump the header into the settings
    settings.update(d_sweep, ignore_errors=True)

    # update the roi_sweep region
    roi_sweep.setRegion((settings['sweep/x_start'], settings['sweep/x_stop']))

# actually overwrite the existing function
d_sweep.after_load_file = d_sweep_after_load








# overwrite the existing shutdown / destroy sequence
def shutdown():
    
    # Stops the sweep before closing
    b_sweep.set_checked(False)
    return
w.event_close = shutdown

if __name__ == '__main__':
    # show the window!
    w.show(True)


import os              as _os

import spinmob as _s
_qtw    = _s._qtw
_qtcore = _s._qtc



def save(filters='*.*', text='Save THIS, facehead!', default_directory='default_directory', force_extension=None):
    """
    Pops up a save dialog and returns the string path of the selected file.
    
    Parameters
    ----------
    filters='*.*'
        Which file types should appear in the dialog.
    text='Save THIS, facehead!'
        Title text for the dialog.
    default_directory='default_directory'
        Key for the spinmob.settings default directory. If you use a name, e.g.
        'my_defaultypoo', for one call of this function, the next time you use
        the same name, it will start in the last dialog's directory by default.
    force_extension=None
        Setting this to a string, e.g. 'txt', will enforce that the filename 
        will have this extension. You can also set this to True, which sets it to 
        the filters value.
    
    Returns
    -------
    Saved path string.
    """
    original_filters = filters
    
    # make sure the filters contains "*.*" as an option!
    if not '*' in filters.split(';'): filters = filters + ";;All files (*)"
    
    # if this type of pref doesn't exist, we need to make a new one
    if default_directory in _settings.keys(): default = _settings[default_directory]
    else:                                     default = ""
    
    # pop up the dialog
    result = _qtw.QFileDialog.getSaveFileName(None,text,default,filters)
    
    # If Qt5, take the zeroth element
    if _s._qt.VERSION_INFO[0:5] == "PyQt5": result = result[0]    
    
    # If the string is empty, quit out.
    if result == '': return None
    
    # Make sure it's a string
    result = str(result)
    
    # Enforce the extension if necessary
    if not force_extension == None:
        
        # If we specify True, default to the filters
        if force_extension == True: force_extension = original_filters
        
        # In case the user put "*.txt" instead of just "txt"
        force_extension = force_extension.replace('*','').replace('.','')

        # If it doesn't end with the right extension, add this.
        if not _os.path.splitext(result)[-1][1:] == force_extension:
            result = result + '.' + force_extension
    
    # Update the user settings.
    _settings[default_directory] = _os.path.split(result)[0]
    return result
    


def load(filters="*.*", text='Select a file, FACEFACE!', default_directory='default_directory'):
    """
    Pops up a dialog for opening a single file. Returns a string path or None.
    
    Parameters
    ----------
    filters='*.*' : str
        String filter for dialog.
    text='Select a file, FACEFACE!' : str
        Dialog window title.
    default_directory='default_directory' : str
        Key to use in spinmob.settings for remembering which directory to start in.
    
    Returns
    -------
    Selected path.
    """
    # make sure the filters contains "*.*" as an option!
    if not '*' in filters.split(';'): filters = filters + ";;All files (*)"
    
    # if this type of pref doesn't exist, we need to make a new one
    if default_directory in _settings.keys(): default = _settings[default_directory]
    else:                                     default = ""
    
    # pop up the dialog
    result = _qtw.QFileDialog.getOpenFileName(None,text,default,filters)
    
    # If Qt5, take the zeroth element
    if _s._qt.VERSION_INFO[0:5] == "PyQt5": result = result[0]
    
    # Make sure it's a string
    result = str(result)    
    
    if result == '': return None
    else:
        _settings[default_directory] = _os.path.split(result)[0]
        return result


def load_multiple(filters="*.*", text='Select some files, FACEFACE!', default_directory='default_directory'):
    """
    Pops up a dialog for opening more than one file. Returns a list of string paths or None.
    
    Parameters
    ----------
    filters='*.*' : str
        String filter for dialog.
    text='Select some files, FACEFACE!' : str
        Dialog window title.
    default_directory='default_directory' : str
        Key to use in spinmob.settings for remembering which directory to start in.
    
    Returns
    -------
    List of selected paths.
    """
    # make sure the filters contains "*.*" as an option!
    if not '*' in filters.split(';'): filters = filters + ";;All files (*)"
    
    # if this type of pref doesn't exist, we need to make a new one
    if default_directory in _settings.keys(): default = _settings[default_directory]
    else:                                     default = ""
    
    # pop up the dialog
    results = _qtw.QFileDialog.getOpenFileNames(None,text,default,filters)
    
    # If Qt5, take the zeroth element
    if _s._qt.VERSION_INFO[0:5] == "PyQt5": results = results[0]
    
    # Make sure it's a string
    result = []
    for r in results: result.append(str(r))       
    
    if len(result)==0: return
    else:
        _settings[default_directory] = _os.path.split(result[0])[0]
        return result


def select_directory(text='Select a directory, POCKETPANTS!', default_directory='default_directory'):
    """
    

    Parameters
    ----------
    text='Select a directory, POCKETPANTS!': str
        Dialog window title text.
    default_directory='default_directory' : str
        Key to use in spinmob.settings for remembering which directory to start in.
    
    Returns
    -------
    result : str
        Selected path.

    """
    # if this type of pref doesn't exist, we need to make a new one
    if default_directory in _settings.keys(): default = _settings[default_directory]
    else:                                     default = ""
    
    # pop up the dialog
    result = _qtw.QFileDialog.getExistingDirectory(None,text,default)
    
    # Make sure it's a string
    result = str(result)        
    
    if result == '': return None
    else:
        _settings[default_directory] = _os.path.split(result)[0]
        return result 




if __name__=='__main__': 
    import spinmob as _s
    _settings = _s.settings
    print(save('*.xyz', force_extension='pdf'))





# -*- coding: utf-8 -*-
import os       as _os
import spinmob  as _s
import unittest as _ut
import shutil   as _sh

import spinmob.egg as _e
import numpy as _n
_g = _e.gui
        

x = None

class Test_egg(_ut.TestCase):
    """
    Test class for databox.
    """

    def setUp(self):
        """
        """
        self.module_path = _os.path.dirname(_s.__file__)
        
        # Path to the data files
        self.data_path = _os.path.join(_os.path.dirname(_s.__file__), 'tests', 'fixtures')

        return

    def tearDown(self):
        """
        """
        return
    
    def test_example(self):
        """
        Just opens a full-featured example and lets the user play with it
        """
        if _os.path.exists('egg_settings'): _sh.rmtree('egg_settings')
        
        import spinmob.egg.example_sweeper as sweeper
        sweeper.d_sweep.load_file(_os.path.join(self.data_path, 'difficult.binary'))
        sweeper.w.show(True)
        
    def test_TreeDictionary(self):
        """
        Runs through a bunch of value setting and getting, to make sure
        the interpretation of different types is reliable.
        """
        
        # clear out the gui settings
        if _os.path.exists('egg_settings'): _sh.rmtree('egg_settings')
        
        test_list_values = [42, 'test', 'pants', 37.2, dict(stuff='otherstuff')]
        test_list_values_stringified = [str(42), 'test', 'pants', str(37.2), str(dict(stuff='otherstuff'))]
        
        # Create the TreeDictionary
        self.d = _e.gui.TreeDictionary('pants.txt')
        
        # Make sure the autosettings path is right.
        self.assertEqual(self.d._autosettings_path, 'pants.txt')
        
        # Create some different value types
        self.d.add_button('button')
        self.d.add_parameter('booly',   value=False)
        self.d.add_parameter('inty',    value=42)
        self.d.add_parameter('floaty',  value=42.0)
        self.d.add_parameter('stringy', value=574,    type='str')
        self.d.add_parameter('listy',   value='test', type='list', values=test_list_values)
        
        # Make sure the get-value returns the right types
        
        # Bool stuff
        self.assertEqual(type(self.d['booly']), bool)
        self.d['booly'] = True
        self.assertEqual(self.d['booly'], True)
        self.d['booly'] = False
        self.assertEqual(self.d['booly'], False)
        
        # Integer stuff
        self.assertEqual(type(self.d['inty']), int)
        self.d['inty'] = '32'
        self.assertEqual(self.d['inty'], 32)
        
        # Float stuff
        self.assertEqual(type(self.d['floaty']), float)
        self.d['floaty'] = '45.5555'
        self.assertEqual(self.d['floaty'], 45.5555)
        
        # See if it's autosaving
        self.assertTrue(_os.path.exists('egg_settings/pants.txt'))
        
        # String stuff
        self.assertEqual(type(self.d['stringy']), str)
        self.d['stringy'] = 47.2
        self.assertEqual(type(self.d['stringy']), str)
        
        # List
        self.assertEqual(self.d.get_list_values('listy'), test_list_values_stringified)
        self.assertEqual(type(self.d['listy']), str)
        self.d['listy'] = 37.2
        self.assertEqual(self.d['listy'], str(37.2))
        
        # Save, load, and make sure the values are still the same (with types)
        self.d.save()
        self.d.load()
        self.assertTrue(_os.path.exists('egg_settings/pants.txt'))
        self.assertEqual(self.d['inty'], 32)
        self.assertEqual(self.d['floaty'], 45.5555)
        self.assertEqual(type(self.d['stringy']),  str)
        self.assertEqual(self.d['listy'], str(37.2))
        
        # Now change a parameter, which should autosave
        self.d['floaty'] = 32
        
        # Create a new one with the same autosettings path
        self.e = _e.gui.TreeDictionary(autosettings_path='pants.txt')
        
        # Test the lazy load & string recall.
        self.e.add_parameter('floaty', 42.0)
        self.assertEqual(self.e['floaty'],32)
        self.e.add_parameter('stringy', 'test')
        self.assertEqual(self.e['stringy'],'47.2')
        
        # Crash test; also tests self.e.get_dictionary(strip_name=True)
        repr(self.e)
        
        # Test get_dictionary with a name
        
        # Set the name
        self.e.name = 'testname'
        ks, d = self.e.get_dictionary()
        self.assertEqual(ks[0], '/testname/floaty')
        
        # Test update
        d['/testname/floaty'] = 77
        self.e.update(d)
        self.assertEqual(self.e['floaty'], 77)
        self.assertEqual(self.e['/testname/floaty'], 77)
        
        # Output to a databox header
        d = _s.data.databox()
        self.e.send_to_databox_header(d)
        self.assertEqual(d.hkeys[0],'/testname/floaty')
        
        # Test load
        self.e.load()
        self.assertEqual(self.e['/testname/floaty'], 77)
        
        
        
        # Test blocking connect_signal_changed
        self.d.add_parameter('a', 0)
        self.d.add_parameter('b', 0)
        def f(*a): self.d['a'] += 1
        self.d.connect_any_signal_changed(f)
        self.d.connect_signal_changed('a', f)
        self.d.set_value('a', 1, block_all_signals=True)
        self.assertEqual(self.d['a'], 1)
        
    def test_DataboxPlot_DataboxProcessor(self):
        
        # Create window
        w = _g.Window(autosettings_path='w')
        
        # Create buttons
        b = w.add(_g.Button('Get Fake Data'))
        l = w.add(_g.Button('Loop', checkable=True))
        
        # Create tab area
        w.new_autorow()
        ts = w.add(_g.TabArea(), alignment=0, column_span=10)
        w.set_column_stretch(4)
        
        # Create plotter and processor
        p = ts.add_tab('Data').add(_g.DataboxPlot('*.dat','p', autoscript=4), alignment=0)
        a = ts.add_tab('Processor').add(_g.DataboxProcessor(databox_source=p), alignment=0)
        
        p.button_script.set_checked(True)
        
        # Window close stops acquisition
        def close(): l.set_checked(False)
        w.event_close = close
        
        # Initial data
        p[0] = [1,2,3,4]
        p[1] = [1,2,1,2]
        p[2] = [0.5,0.5,0.5,0.5]
        p[3] = [2,1,2,1]
        p[4] = [0.3,0.1,20,0.3]
        p.plot()
        
        # New Data
        def get_fake_data():
            p.clear()
            p['t']  = _n.linspace(0,10,1000)
            p['V1'] = 0.01*_n.random.normal(size=1000)
            p['V2'] = _n.random.normal(size=1000)
            p.plot()
            a.run()
        b.signal_clicked.connect(get_fake_data)
        
        # Loop
        def loop():
            while l.is_checked():
                get_fake_data()
                w.sleep(0.25)
        l.signal_clicked.connect(loop)
        
        w.show(True)
        global x
        x = p
    
      

if __name__ == "__main__":
    _ut.main()
    
    self = Test_egg()
    


import numpy   as _n
import time    as _t
import spinmob as _s
from spinmob import egg



##### GUI DESIGN

# create the main window
w = egg.gui.Window(autosettings_path='example_daq_w.cfg')

# add the "go" button
b = w.place_object(egg.gui.Button("Pretend to Acquire Data", checkable=True)).set_width(135)
i = w.place_object(egg.gui.NumberBox(int=True)).set_width(50)

# add a tabbed interface for the plotting area, spanning the first and second rows
tabs = w.place_object(egg.gui.TabArea(), row_span=2, alignment=0)

# add a tab for some plots
t1 = tabs.add_tab("Plots")
t2 = tabs.add_tab("Analysis")

# add a databox plotter object to the tab
d1 = t1.place_object(egg.gui.DataboxPlot(autosettings_path='example_daq_p1.cfg'), alignment=0)
d2 = t2.place_object(egg.gui.DataboxPlot(autosettings_path='example_daq_p2.cfg'), alignment=0)

# move to the second row and add a TreeDictionary for
# our "settings"
w.new_autorow()
s = w.place_object(egg.gui.TreeDictionary('example_daq_s.cfg'), column_span=2)



##### DATA OBJECTS

# add various parameters to use later
s.add_parameter("settings/simulated_input/iterations", value=0,            type='int')
s.add_parameter("settings/simulated_input/channels",   value=3,            type='int',   limits=(1, 8))
s.add_parameter("settings/simulated_input/duration",   value=0.02,         type='float', limits=(0.001,None), dec=True, siPrefix=True, suffix='s')
s.add_parameter("settings/simulated_input/points",     value=1000,         type='int',   limits=(2,    None))
s.add_parameter("settings/simulated_input/source",     value="cos(500*t)", type='str')
s.add_parameter("settings/simulated_input/noise",      value=0.2,          type='float')
s.add_parameter("settings/other_stuff/name",           value="moose",      type='str')
s.add_parameter("settings/other_stuff/quality",        value="round",      type='list', values=["round", "faceted", "regular moose"])

# load previous settings if they exist
s.load()



##### FUNCTIONALITY

# define a function to create fake data
# this is where you might put code for
# talking to a daq card or sillyscope
def get_fake_data():

    # try to evaluate the source script
    try:
        # first get all the extra numpy globals
        g = _n.__dict__

        # update these globals with extra stuff needed for evaluation
        g.update(dict(t=d1[0]+s["settings/simulated_input/noise"]*_n.random.rand()))
        y = eval(s["settings/simulated_input/source"], g)+_n.random.random(len(d1['t']))

    # default to zeros
    except:
        print("ERROR: Invalid source script.")
        y = d1[0]*0.0

    # pretend this acquisition actually took time (avoids black holes)
    _t.sleep(s["settings/simulated_input/duration"])

    return y


# define a function to be called whenever the acquire button is pressed
def acquire_button_clicked(*a):

    # don't start another loop if the button is unchecked!
    if not b.is_checked(): return

    # reset the counter
    i.set_value(0)

    print("Sarting fake acquisition loop...")

    # start the loop and keep looping until someone
    # unchecks the acquisition button or we max out the iterations
    # setting iterations = 0 will loop infinitely
    while b.is_checked()                                          \
    and (i.get_value() < s["settings/simulated_input/iterations"] \
         or s["settings/simulated_input/iterations"] == 0):

        # reset the databox
        d1.clear()
        d2.clear()
        
        # all the information to the databox headers
        s.send_to_databox_header(d1)
        s.send_to_databox_header(d2)

        # create the fake time data
        d1['t'] = _n.linspace(0,s["settings/simulated_input/duration"], s["settings/simulated_input/points"])

        # create the fake channel data and do some analysis on it
        for n in range(s["settings/simulated_input/channels"]):

            # get the channel name
            c = 'ch'+str(n)

            # create the data
            d1[c] = get_fake_data()
            
            # get a power spectral density
            d2['f'], d2[c] = _s.fun.psd(d1['t'], d1[c])
            
        # increment the counter
        i.increment()

        # update the plot (note updating d1 updates p1.databox by reference)
        # and autosave (only does anything if this is enabled on the gui)
        d1.plot(); d1.autosave()
        d2.plot(); d2.autosave()
        
        # process other window events so the GUI doesn't freeze
        w.process_events()

    # in case the button is still checked
    b.set_checked(False)

    print("Fake acquisition stopped.")

# connect the button
b.signal_clicked.connect(acquire_button_clicked)

# Define a function to save the settings whenever
# they change and connect the signal
def settings_changed(*a): s.save()

# connecting is a little different for TreeDictionaries
s.connect_any_signal_changed(settings_changed)



# overwrite the existing shutdown / destroy sequence
def shutdown():
    print("Closing but not destroying...")
    b.set_checked(False)
    return
w.event_close = shutdown

# show the window!
w.show(True)


#import wx    as _wx
import numpy   as _n
import os      as _os
import shutil  as _shutil
import spinmob as _s
import pickle  as _cPickle



class averager:
    """
    This object will keep a running average of any "measured" quantity 
    (i.e. a single value drawn from a fluctuating source) along with a running 
    estimate of the standard deviation for individual measurements, and the
    standard error on the mean, as estimated from the supplied measurements 
    themselves.
    
    This object is particularly useful when averaging together large data sets, 
    such as long traces from a triggered oscilloscope or a series of power 
    spectral densities. The algorithm requires a fixed quantity of memory 
    regardless of how many averages one takes. 
    
    As with any averaging technique, make sure each added measurement can be 
    added to the running total without succumbing to round-off errors. This 
    is especially important for low bit depths. For example, 16-bit floating 
    point data runs into roundoff errors after only 3 decimal places (!), 
    meaning you can only effectively include up to 1024 data sets in a 
    calculation of the mean, with artifacts appearing significantly sooner 
    than this. 
    
    This averager also has the option to calculate an exponential moving average
    (i.e., a low-pass filter) by specifying lowpass_frames > 0. This also provides
    an estimate of the moving variance, which is systematically underestimated
    as lowpass_frames is made smaller. At 2, for example, the variance is
    underestimated by ~10%, and at 1, the variance becomes zero.
    
    Arguments
    ---------
    name='a'
        Name of the data set.
    
    lowpass_frames=0
        If zero, averager will compute the average of all supplied data sets. 
        If positive, averager will perform a DSP low-pass on the supplied data
        sets with lowpass_frames as the time constant (units of data sets). 
        Note this algorithm systematically underestimates the variance as
        lowpass_frames approaches 1, and is within 10% of the true value for
        lowpass_frames > 2.                                                               
     
    precision=numpy.float64
        Numpy dtype used in internal calculations. numpy.float32 is usually 
        sufficient, but numpy.float64 is safer if you are not strapped for 
        memory. Major precision issues seem to first appear as numpy.nan's 
        appearing in the calculation of self.variance_mean and 
        self.variance_sample, due to round off error on the 
        subtraction prior to the square root.
    
    ignore_nan=True
        If True, sets all NaN's appearing in the calculation to zero.
    
    Internal Quantities
    -------------------
    self.lowpass_frames
        Time constant for the DSP low-pass (see above).

    self.ignore_nan
        Whether to set NaN's appearing in calculations to zero.
        
    self.mean
        Running mean value of all supplied measurements.
        
    self.mean_squared
        Running mean^2 value of all supplied measurements.
    
    self.name
        Name of the data set.
    
    self.N
        Number of measurements that have been included thus far.
    
    self.precision
        Numpy dtype used in internal calculations. 
        
    self.variance_mean
        Variance about the mean estimated from the supplied measurements.
        Assumes all measurements are independently drawn from the same statistical
        distribution.
        
    self.variance_sample
        Variance on individual measurements. 
        Assumes all measurements are independently drawn from the same statistical
        distribution.
    
    Methods
    -------
    add_measurement(y) 
        Add a new measurement and update the above quantities.
    
    reset()
        Reset the counter and all quantities to zero.
    """
    def __init__(self, name='a', lowpass_frames=0, precision=_n.float64, ignore_nan=True): 
        
        # Don't forget to update the _copy method!
        self.ignore_nan      = ignore_nan
        self.lowpass_frames  = lowpass_frames
        self.name            = name
        self.N               = 0  
        self.mean            = 0  
        self.mean_squared    = 0
        self.precision       = precision
        self.variance_mean   = 0  
        self.variance_sample = 0
        
    def __repr__(self):
        return '<tools.averager '+repr(self.name)+' N='+str(self.N)+'>'
    
    def _copy(self):
        """
        Creates and returns a copy of this instance.
        """
        # Make a new instance
        a = type(self)()
        a.ignore_nan      = self.ignore_nan
        a.lowpass_frames  = self.lowpass_frames
        a.name            = self.name
        a.N               = self.N
        a.mean            = self.mean
        a.mean_squared    = self.mean_squared
        a.precision       = self.precision
        a.variance_mean   = self.variance_mean
        a.variance_sample = self.variance_sample
        return a
    
    def add(self, y):
        """
        Add a new measurement to the pool, updating the counter, means, and 
        standard deviations. Note this modifies the instance and returns it.
        
        We use a modified Welford's online algorithm because it's more stable against
        catastrophic cancellation.
        
        Parameters
        ----------
        y:
            Number, numpy array, or other object (anything that can be summed
            multiplied and divided in a meaningful way) representing a single 
            measurement. This will be automatically converted to the format 
            specified by self.precision.
        """
        
        # Make sure the shape of y matches with the existing data, or reset
        if not _n.shape(y) == _n.shape(self.mean): self.reset()

        # Convert to the desired precision.
        y = self.precision(y)

        # Update the counter
        self.N += 1
        
        # if it is our first data set, initialize.
        if self.N == 1:
            self.mean            = y
            self.mean_squared    = y*y
            
            if _s.fun.is_iterable(y):
                self.variance_mean   = _n.zeros_like(y)
                self.variance_sample = _n.zeros_like(y)
            else:
                self.variance_mean   = 0
                self.variance_sample = 0
        
        # If we're doing a cumulative mean and variance N>1
        elif not self.lowpass_frames:
            N = self.N
            
            # Remember the previous mean
            old_mean         = self.mean
            old_mean_squared = self.mean_squared
            
            # Get the new mean
            self.mean         += (y   - self.mean        ) / N
            self.mean_squared += (y*y - self.mean_squared) / N
            
            # Calculate the variances
            self.variance_sample = (self.mean_squared - self.mean*self.mean)*self.N/(self.N-1)
            #self.variance_sample = self.variance_sample*(N-2)/(N-1) + (y-old_mean)*(y-self.mean)/(N-1)
            self.variance_mean   = self.variance_sample/self.N
            
        # Otherwise we're doing DSP low-pass N>1
        else:
            
            # Time constant shorthand (with a hack to make the initial variance more responsive)
            tau   = min(self.lowpass_frames, self.N)
            kappa = 1.0/tau
            
            # Remember the previous value
            old_mean = self.mean
            
            # Update the mean & variance_sample with DSP low-pass
            self.mean         = kappa*y   + (1-kappa)*old_mean
            self.mean_squared = kappa*y*y + (1-kappa)*self.mean_squared
            
            # Get the standard error on the mean
            ek  = _n.exp(kappa)
            emk = 1/ek
            self.variance_sample = 0.5*(1+ek)*(self.mean_squared - self.mean*self.mean)
            self.variance_mean = self.variance_sample * (1-_n.exp(-kappa))**2 / (1-_n.exp(-2*kappa))
            
        
        # If we are ignoring the nan's, set them to zero.
        if self.ignore_nan: 
            if type(self.variance_sample) == _n.ndarray: _n.nan_to_num(self.variance_sample, copy=False)
            else:   self.variance_sample = _n.nan_to_num(self.variance_sample)
            
            if type(self.variance_mean) == _n.ndarray: _n.nan_to_num(self.variance_mean, copy=False)
            else:   self.variance_mean = _n.nan_to_num(self.variance_mean)
            
        # Otherwise, print a warning but keep moving.
        elif _n.isnan(self.variance_sample).any() or _n.isnan(self.variance_mean).any(): 
            print("WARNING: "+repr(self)+" Some elements of self.variance_mean or self.variance_sample are NaN.")
        
        return self
    
    def __add__(self, y):
        """
        Creates a copy of this instance, adds runs add(y) on the copy, and 
        returns the copy.
        
        Parameters
        ----------
        y:
            To be sent to the copy's add() function.
        """
        c = self._copy()
        return c.add(y)
    
    def reset(self): 
        """
        Resets the average values and counter to zero.
        """
        self.__init__(name           = self.name,
                      lowpass_frames = self.lowpass_frames,
                      precision      = self.precision, 
                      ignore_nan     = self.ignore_nan)
        return self
        



def coarsen_array(a, level=2, method='mean'):
    """
    Returns a coarsened (binned) version of the data. Currently supports
    any of the numpy array operations, e.g. min, max, mean, std, ...
    
    level=2 means every two data points will be binned.
    level=0 or 1 just returns a copy of the array
    """
    if a is None: return None    
    
    # make sure it's a numpy array, and that we don't destroy the original.
    a = _n.array(a)
    
    # Make sure it's an integer!
    level=int(level)
    
    # quickest option
    if level in [0,1,False]: return a

    # otherwise assemble the python code to execute
    code = 'a.reshape(-1, level).'+method+'(axis=1)'

    # execute, making sure the array can be reshaped!
    try: return eval(code, dict(a=a[0:int(len(a)/level)*level], level=level))
    except:
        print("ERROR: Could not coarsen array with method "+repr(method))
        return a

def coarsen_data(x, y, ey=None, ex=None, level=2, exponential=False):
    """
    Coarsens the supplied data set. Returns coarsened arrays of x, y, along with
    quadrature-coarsened arrays of ey and ex if specified.
    
    Parameters
    ----------
    x, y
        Data arrays. Can be lists (will convert to numpy arrays).
        These are coarsened by taking an average.
    ey=None, ex=None
        y and x uncertainties. Accepts arrays, lists, or numbers. 
        These are coarsened by averaging in quadrature.
    level=2
        For linear coarsening (default, see below), every n=level points will
        be averaged together (in quadrature for errors). For exponential
        coarsening, bins will be spaced by the specified scaling=level factor;
        for example, level=1.4 will group points within 40% of each other's x
        values. This is a great option for log-x plots, as the outcome will be
        evenly spaced.
    exponential=False
        If False, coarsen using linear spacing. If True, the bins will be 
        exponentially spaced by the specified level. 
    """
    
    # Normal coarsening
    if not exponential:
        
        # Coarsen the data
        xc  = coarsen_array(x, level, 'mean')
        yc  = coarsen_array(y, level, 'mean')
        
        # Coarsen the y error in quadrature
        if not ey is None:
            if not is_iterable(ey): ey = [ey]*len(y)
            eyc = _n.sqrt(coarsen_array(_n.power(ey,2)/level, level, 'mean'))
        
        # Coarsen the x error in quadrature
        if not ex is None:
            if not is_iterable(ey): ex = [ex]*len(x)
            exc = _n.sqrt(coarsen_array(_n.power(ex,2)/level, level, 'mean'))
        
    # Exponential coarsen    
    else:
        # Make sure the data are arrays
        x = _n.array(x)
        y = _n.array(y)
        
        # Create the new arrays to fill
        xc = []
        yc = []
        if not ey is None: 
            if not is_iterable(ey): ey = _n.array([ey]*len(y))
            eyc = []
        if not ex is None:
            if not is_iterable(ex): ex = _n.array([ex]*len(x))
            exc = []
        
        # Find the first element that is greater than zero    
        x0 = x[x>0][0]
        
        # Now loop over the exponential bins
        n  = 0
        while x0*level**n < x[-1]:
            
            # Get all the points between x[n] and x[n]*r
            mask = _n.logical_and(x0*level**n <= x, x < x0*level**(n+1))
                    
            # Only do something if points exist from this range!
            if len(x[mask]):
                
                # Take the average x value
                xc.append(_n.average(x[mask]))
                yc.append(_n.average(y[mask]))
            
                # do the errors in quadrature
                if not ey is None: eyc.append(_n.sqrt(_n.average((ey**2)[mask])/len(ey[mask])))
                if not ex is None: exc.append(_n.sqrt(_n.average((ex**2)[mask])/len(ex[mask])))
            
            # Increment the counter
            n += 1
        
        # Done exponential loop

    # Done coarsening        
    # Return depending on situation
    if   ey is None and ex is None: return _n.array(xc), _n.array(yc)
    elif ex is None               : return _n.array(xc), _n.array(yc), _n.array(eyc)
    elif ey is None               : return _n.array(xc), _n.array(yc), _n.array(exc)
    else                          : return _n.array(xc), _n.array(yc), _n.array(eyc), _n.array(exc)
        
        

def coarsen_matrix(Z, xlevel=0, ylevel=0, method='average'):
    """
    This returns a coarsened numpy matrix.

    method can be 'average', 'maximum', or 'minimum'
    """

    # coarsen x
    if not ylevel:
        Z_coarsened = Z
    else:
        temp = []
        for z in Z: temp.append(coarsen_array(z, ylevel, method))
        Z_coarsened = _n.array(temp)

    # coarsen y
    if xlevel:
        Z_coarsened = Z_coarsened.transpose()
        temp = []
        for z in Z_coarsened: temp.append(coarsen_array(z, xlevel, method))
        Z_coarsened = _n.array(temp).transpose()

    return Z_coarsened



    # first coarsen the columns (if necessary)
    if ylevel:
        Z_ycoarsened = []
        for c in Z: Z_ycoarsened.append(coarsen_array(c, ylevel, method))
        Z_ycoarsened = _n.array(Z_ycoarsened)

    # now coarsen the rows
    if xlevel: return coarsen_array(Z_ycoarsened, xlevel, method)
    else:      return _n.array(Z_ycoarsened)


def erange(start, end, steps):
    """
    Returns a numpy array over the specified range taking geometric steps.

    See also numpy.logspace()
    """
    if start == 0:
        print("Nothing you multiply zero by gives you anything but zero. Try picking something small.")
        return None
    if end == 0:
        print("It takes an infinite number of steps to get to zero. Try a small number?")
        return None

    # figure out our multiplication scale
    x = (1.0*end/start)**(1.0/(steps-1))

    # now generate the array
    ns = _n.array(list(range(0,steps)))
    a =  start*_n.power(x,ns)

    # tidy up the last element (there's often roundoff error)
    a[-1] = end

    return a

def is_a_number(s):
    """
    This takes an object and determines whether it's a number or a string
    representing a number.
    """
    if _s.fun.is_iterable(s) and not type(s) == str: return False

    try:
        float(s)
        return 1
    except:
        try: 
            complex(s)
            return 2
        except:            
            try: 
                complex(s.replace('(','').replace(')','').replace('i','j'))
                return 2
            except:            
                return False

def is_iterable(a):
    """
    Determine whether the object is iterable, but not a string.
    This function is left over from a time when Python was 2, not 3.
    """
    return hasattr(a, '__iter__') and not type(a) == str


def append_to_file(path, string):
    file = open(path, 'a')
    file.write(string)
    file.close()

def array_shift(a, n, fill="average"):
    """
    This will return an array with all the elements shifted forward in index by n.

    a is the array
    n is the amount by which to shift (can be positive or negative)

    fill="average"      fill the new empty elements with the average of the array
    fill="wrap"         fill the new empty elements with the lopped-off elements
    fill=37.2           fill the new empty elements with the value 37.2
    """

    new_a = _n.array(a)

    if n==0: return new_a

    fill_array = _n.array([])
    fill_array.resize(_n.abs(n))

    # fill up the fill array before we do the shift
    if   fill is "average": fill_array = 0.0*fill_array + _n.average(a)
    elif fill is "wrap" and n >= 0:
        for i in range(0,n): fill_array[i] = a[i-n]
    elif fill is "wrap" and n < 0:
        for i in range(0,-n): fill_array[i] = a[i]
    else:   fill_array = 0.0*fill_array + fill

    # shift and fill
    if n > 0:
        for i in range(n, len(a)): new_a[i] = a[i-n]
        for i in range(0, n):      new_a[i] = fill_array[i]
    else:
        for i in range(0, len(a)+n): new_a[i] = a[i-n]
        for i in range(0, -n):       new_a[-i-1] = fill_array[-i-1]

    return new_a



def assemble_covariance(error, correlation):
    """
    This takes an error vector and a correlation matrix and assembles the covariance
    """

    covariance = []
    for n in range(0, len(error)):
        covariance.append([])
        for m in range(0, len(error)):
            covariance[n].append(correlation[n][m]*error[n]*error[m])
    return _n.array(covariance)






def chi_squared(p, f, xdata, ydata):
    return(sum( (ydata - f(p,xdata))**2 ))








def combine_dictionaries(a, b):
    """
    returns the combined dictionary.  a's values preferentially chosen
    """

    c = {}
    for key in list(b.keys()): c[key]=b[key]
    for key in list(a.keys()): c[key]=a[key]
    return c

def data_from_file(path, delimiter=" "):
    lines = read_lines(path)
    x = []
    y = []
    for line in lines:
       s=line.split(delimiter)
       if len(s) > 1:
           x.append(float(s[0]))
           y.append(float(s[1]))
    return([_n.array(x), _n.array(y)])


def data_to_file(path, xarray, yarray, delimiter=" ", mode="w"):
    file = open(path, mode)
    for n in range(0, len(xarray)):
        file.write(str(xarray[n]) + delimiter + str(yarray[n]) + '\n')
    file.close()









def decompose_covariance(c):
    """
    This decomposes a covariance matrix into an error vector and a correlation matrix
    """

    # make it a kickass copy of the original
    c = _n.array(c)

    # first get the error vector
    e = []
    for n in range(0, len(c[0])): e.append(_n.sqrt(c[n][n]))

    # now cycle through the matrix, dividing by e[1]*e[2]
    for n in range(0, len(c[0])):
        for m in range(0, len(c[0])):
            c[n][m] = c[n][m] / (e[n]*e[m])

    return [_n.array(e), _n.array(c)]

def derivative(xdata, ydata):
    """
    performs d(ydata)/d(xdata) with nearest-neighbor slopes
    must be well-ordered, returns new arrays [xdata, dydx_data]

    neighbors:
    """
    D_ydata = []
    D_xdata = []
    for n in range(1, len(xdata)-1):
        D_xdata.append(xdata[n])
        D_ydata.append((ydata[n+1]-ydata[n-1])/(xdata[n+1]-xdata[n-1]))

    return [D_xdata, D_ydata]

def derivative_fit(xdata, ydata, neighbors=1):
    """
    loops over the data points, performing a least-squares linear fit of the
    nearest neighbors at each point. Returns an array of x-values and slopes.

    xdata should probably be well-ordered.

    neighbors   How many data point on the left and right to include.
    """

    x    = []
    dydx = []
    nmax = len(xdata)-1

    for n in range(nmax+1):
        # get the indices of the data to fit
        i1 = max(0, n-neighbors)
        i2 = min(nmax, n+neighbors)

        # get the sub data to fit
        xmini = _n.array(xdata[i1:i2+1])
        ymini = _n.array(ydata[i1:i2+1])

        slope, intercept = fit_linear(xmini, ymini)

        # make x the average of the xmini
        x.append(float(sum(xmini))/len(xmini))
        dydx.append(slope)

    return _n.array(x), _n.array(dydx)

def difference(ydata1, ydata2):
    """

    Returns the number you should add to ydata1 to make it line up with ydata2

    """

    y1 = _n.array(ydata1)
    y2 = _n.array(ydata2)

    return(sum(y2-y1)/len(ydata1))




def distort_matrix_X(Z, X, f, new_xmin, new_xmax, subsample=3):
    """
    Applies a distortion (remapping) to the matrix Z (and x-values X) using function f.
    returns new_Z, new_X

    f is an INVERSE function old_x(new_x)

    Z is a matrix. X is an array where X[n] is the x-value associated with the array Z[n].

    new_xmin, new_xmax is the possible range of the distorted x-variable for generating Z

    points is how many elements the stretched Z should have. "auto" means use the same number of bins
    """

    Z = _n.array(Z)
    X = _n.array(X)
    points = len(Z)*subsample


    # define a function for searching
    def zero_me(new_x): return f(new_x)-target_old_x

    # do a simple search to find the new_x that gives old_x = min(X)
    target_old_x = min(X)
    new_xmin = find_zero_bisect(zero_me, new_xmin, new_xmax, _n.abs(new_xmax-new_xmin)*0.0001)
    target_old_x = max(X)
    new_xmax = find_zero_bisect(zero_me, new_xmin, new_xmax, _n.abs(new_xmax-new_xmin)*0.0001)

    # now loop over all the new x values
    new_X = []
    new_Z = []
    bin_width = float(new_xmax-new_xmin)/(points)
    for new_x in frange(new_xmin, new_xmax, bin_width):

        # make sure we're in the range of X
        if f(new_x) <= max(X) and f(new_x) >= min(X):

            # add this guy to the array
            new_X.append(new_x)

            # get the interpolated column
            new_Z.append( interpolate(X,Z,f(new_x)) )

    return _n.array(new_Z), _n.array(new_X)




def distort_matrix_Y(Z, Y, f, new_ymin, new_ymax, subsample=3):
    """
    Applies a distortion (remapping) to the matrix Z (and y-values Y) using function f.
    returns new_Z, new_Y

    f is a function old_y(new_y)

    Z is a matrix. Y is an array where Y[n] is the y-value associated with the array Z[:,n].

    new_ymin, new_ymax is the range of the distorted x-variable for generating Z

    points is how many elements the stretched Z should have. "auto" means use the same number of bins
    """

    # just use the same methodology as before by transposing, distorting X, then
    # transposing back
    new_Z, new_Y = distort_matrix_X(Z.transpose(), Y, f, new_ymin, new_ymax, subsample)
    return new_Z.transpose(), new_Y





def dumbguy_minimize(f, xmin, xmax, xstep):
    """
    This just steps x and looks for a peak

    returns x, f(x)
    """

    prev = f(xmin)
    this = f(xmin+xstep)
    for x in frange(xmin+xstep,xmax,xstep):
        next = f(x+xstep)

        # see if we're on top
        if this < prev and this < next: return x, this

        prev = this
        this = next

    return x, this

def elements_are_numbers(array):
    """
    Tests whether the elements of the supplied array are numbers.
    """

    # empty case
    if len(array) == 0: return 0

    output_value = 1
    for x in array:

        # test it and die if it's not a number
        test = is_a_number(x)
        if not test: return False

        # mention if it's complex
        output_value = max(output_value,test)
    
    return output_value

def elements_are_strings(array, start_index=0, end_index=-1):
    if len(array) == 0: return 0

    if end_index < 0: end_index=len(array)-1

    for n in range(start_index, end_index+1):
        if not type(array[n]) == str: return 0
    return 1

def elements_are_iterable(array):
    """
    Returns True if every element is a list/array-like object (not a string).
    """
    for a in array: 
        if not is_iterable(a): 
            return False
    return True
        
def equalize_list_lengths(a,b):
    """
    Modifies the length of list a to match b. Returns a. 
    a can also not be a list (will convert it to one).
    a will not be modified.
    """
    if not _s.fun.is_iterable(a): a = [a]
    a = list(a)
    while len(a)>len(b): a.pop(-1)
    while len(a)<len(b): a.append(a[-1])
    return a

def find_N_peaks(array, N=4, max_iterations=100, rec_max_iterations=3, recursion=1):
    """
    This will run the find_peaks algorythm, adjusting the baseline until exactly N peaks are found.
    """

    if recursion<0: return None

    # get an initial guess as to the baseline
    ymin = min(array)
    ymax = max(array)

    for n in range(max_iterations):

        # bisect the range to estimate the baseline
        y1 = (ymin+ymax)/2.0

        # now see how many peaks this finds. p could have 40 for all we know
        p, s, i = find_peaks(array, y1, True)

        # now loop over the subarrays and make sure there aren't two peaks in any of them
        for n in range(len(i)):
            # search the subarray for two peaks, iterating 3 times (75% selectivity)
            p2 = find_N_peaks(s[n], 2, rec_max_iterations, rec_max_iterations=rec_max_iterations, recursion=recursion-1)

            # if we found a double-peak
            if not p2 is None:
                # push these non-duplicate values into the master array
                for x in p2:
                    # if this point is not already in p, push it on
                    if not x in p: p.append(x+i[n]) # don't forget the offset, since subarrays start at 0


        # if we nailed it, finish up
        if len(p) == N: return p

        # if we have too many peaks, we need to increase the baseline
        if len(p) > N: ymin = y1

        # too few? decrease the baseline
        else:          ymax = y1

    return None


def find_peaks(array, baseline=0.1, return_subarrays=False):
    """
    This will try to identify the indices of the peaks in array, returning a list of indices in ascending order.

    Runs along the data set until it jumps above baseline. Then it considers all the subsequent data above the baseline
    as part of the peak, and records the maximum of this data as one peak value.
    """

    peaks = []

    if return_subarrays:
        subarray_values  = []
        subarray_indices = []

    # loop over the data
    n = 0
    while n < len(array):
        # see if we're above baseline, then start the "we're in a peak" loop
        if array[n] > baseline:

            # start keeping track of the subarray here
            if return_subarrays:
                subarray_values.append([])
                subarray_indices.append(n)

            # find the max
            ymax=baseline
            nmax = n
            while n < len(array) and array[n] > baseline:
                # add this value to the subarray
                if return_subarrays:
                    subarray_values[-1].append(array[n])

                if array[n] > ymax:
                    ymax = array[n]
                    nmax = n

                n = n+1

            # store the max
            peaks.append(nmax)

        else: n = n+1

    if return_subarrays: return peaks, subarray_values, subarray_indices
    else:                return peaks


def find_two_peaks(data, remove_background=True):
    """

    Returns two indicies for the two maxima

    """

    y  = _n.array( data            )
    x  = _n.array( list(range(0,len(y))) )

    # if we're supposed to, remove the linear background
    if remove_background:
        [slope, offset] = fit_linear(x,y)
        y = y - slope*x
        y = y - min(y)

    # find the global maximum
    max1   = max(y)
    n1     = index(max1, y)

    # now starting at n1, work yourway left and right until you find
    # the left and right until the data drops below a 1/2 the max.
    # the first side to do this gives us the 1/2 width.
    np = n1+1
    nm = n1-1
    yp = max1
    ym = max1
    width = 0
    while 0 < np < len(y) and 0 < nm < len(y):
        yp = y[np]
        ym = y[nm]

        if yp <= 0.5*max1 or ym <= 0.5*max1:
            width = np - n1
            break

        np += 1
        nm -= 1



    # if we didn't find it, we pooped out
    if width == 0:
        return [n1,-1]

    # this means we have a valid 1/2 width.  Find the other max in the
    # remaining data
    n2 = nm
    while 1 < np < len(y)-1 and 1 < nm < len(y)-1:
        if y[np] > y[n2]:
            n2 = np
        if y[nm] > y[n2]:
            n2 = nm
        np += 1
        nm -= 1

    return([n1,n2])





def find_zero_bisect(f, xmin, xmax, xprecision):
    """
    This will bisect the range and zero in on zero.
    """
    if f(xmax)*f(xmin) > 0:
        print("find_zero_bisect(): no zero on the range",xmin,"to",xmax)
        return None

    temp = min(xmin,xmax)
    xmax = max(xmin,xmax)
    xmin = temp

    xmid = (xmin+xmax)*0.5
    while xmax-xmin > xprecision:
        y = f(xmid)

        # pick the direction with one guy above and one guy below zero
        if y > 0:
            # move left or right?
            if f(xmin) < 0: xmax=xmid
            else:           xmin=xmid

        # f(xmid) is below zero
        elif y < 0:
            # move left or right?
            if f(xmin) > 0: xmax=xmid
            else:           xmin=xmid

        # yeah, right
        else: return xmid

        # bisect again
        xmid = (xmin+xmax)*0.5

    return xmid


def fit_linear(xdata, ydata):
    """
    Returns slope and intercept of line of best fit:
        
        y = a*x + b
    
    through the supplied data. Returns a,b.
    
    Parameters
    ----------
    xdata, ydata:
        Arrays of x data and y data (having matching lengths).

    """
    x = _n.array(xdata)
    y = _n.array(ydata)

    ax  = _n.average(x)
    ay  = _n.average(y)
    axx = _n.average(x*x)
    ayx = _n.average(y*x)

    slope     = (ayx - ay*ax) / (axx - ax*ax)
    intercept = ay - slope*ax

    return slope, intercept






def frange(start, end, inc=1.0):
    """
    A range function, that accepts float increments and reversed direction.

    See also numpy.linspace()
    """

    start = 1.0*start
    end   = 1.0*end
    inc   = 1.0*inc

    # if we got a dumb increment
    if not inc: return _n.array([start,end])

    # if the increment is going the wrong direction
    if 1.0*(end-start)/inc < 0.0:
        inc = -inc

    # get the integer steps
    ns = _n.array(list(range(0, int(1.0*(end-start)/inc)+1)))

    return start + ns*inc


def generate_fake_data(f='2*x-5', x=_n.linspace(-5,5,11), ey=1, ex=0, include_errors=False, **kwargs):
    """
    Generates a set of fake data from the underlying "reality" (or mean
    behavior) function f.
    
    Parameters
    ----------
    f='2*x-5':
        Underlying "reality" function or mean behavior. This can be any
        python-evaluable string, and will have access to all the numpy
        functions (e.g., cos), scipy's special functions (e.g., erf), and
        any other variables defined by keyword arguments
    x=numpy.linspace(-5,5,11):
        x-values to evaluate.
    ey=1, ex=0:
        Uncertainty "strength" for x and y data. This can be a constant or an 
        array of values. If the distributions (below) are normal, this 
        corresponds to the standard deviation.
    include_errors=True
        Whether the databox should include a column for ex and ey.
    
    Keyword arguments are used as additional globals in the function evaluation.
    
    Returns a databox containing the data and other relevant information in the
    header.
    """
    
    # Make a fitter object, which handily interprets string functions
    # The "+0*x" is a trick to ensure the function takes x as an argument 
    # (makes it a little more idiot proof).
    fitty = _s.data.fitter().set_functions(f+"+0*x",'') 
    
    # Make sure both errors are arrays of the right length
    if not _s.fun.is_iterable(ex): ex = _n.array([ex]*len(x))
    if not _s.fun.is_iterable(ey): ey = _n.array([ey]*len(x))
    
    # Get the x and y exact values first, then randomize
    x = _n.array(x) 
    y = fitty.f[0](x)
    
    x = _n.random.normal(_n.array(x),ex)
    y = _n.random.normal(y,          ey)
    
    # make a databox
    d = _s.data.databox()
    d['x']  = x
    d['y']  = y
    
    if include_errors:
        d['ey'] = ey
        d['ex'] = ex
    
    d.h(reality=f, ey=ey[0], ex=ex[0])
    
    return d



def get_shell_history():
    """
    This only works with some shells.
    """
    # try for ipython
    if 'get_ipython' in globals():
        a = list(get_ipython().history_manager.input_hist_raw)
        a.reverse()
        return a

    elif 'SPYDER_SHELL_ID' in _os.environ:
        try:
            p = _os.path.join(_settings.path_user, ".spyder2", "history.py")
            a = read_lines(p)
            a.reverse()
            return a
        except:
            pass

    # otherwise try pyshell or pycrust (requires wx)
    else:
        try:
            import wx
            for x in wx.GetTopLevelWindows():
                if type(x) in [wx.py.shell.ShellFrame, wx.py.crust.CrustFrame]:
                    a = x.shell.GetText().split(">>>")
                    a.reverse()
                    return a
        except:
            pass

    return ['shell history not available']


def imax(array):
    """
    Returns the index of the maximum of array.
    """
    return index(max(array), array)

def imin(array):
    """
    Returns the index of the minimum of array.
    """
    return index(min(array), array)

def index(value, array):
    """
    Array search that behaves like I want it to. Totally dumb, I know.
    """
    i = array.searchsorted(value)
    if i == len(array): return -1
    else:               return i

def index_nearest(value, array):
    """
    expects a _n.array
    returns the global minimum of (value-array)^2
    """

    a = (array-value)**2
    return index(a.min(), a)

def index_next_crossing(value, array, starting_index=0, direction=1):
    """
    starts at starting_index, and walks through the array until
    it finds a crossing point with value

    set direction=-1 for down crossing
    """

    for n in range(starting_index, len(array)-1):
        if  (value-array[n]  )*direction >= 0         \
        and (value-array[n+1])*direction <  0: return n

    # no crossing found
    return -1




def insert_ordered(value, array):
    """
    This will insert the value into the array, keeping it sorted, and returning the
    index where it was inserted
    """

    index = 0

    # search for the last array item that value is larger than
    for n in range(0,len(array)):
        if value >= array[n]: index = n+1

    array.insert(index, value)
    return index







def integrate_data(xdata, ydata, xmin=None, xmax=None, autozero=0):
    """
    Numerically integrates up the ydata using the trapezoid approximation.
    estimate the bin width (scaled by the specified amount).
    Returns (xdata, integrated ydata).

    autozero is the number of data points to use as an estimate of the background
    (then subtracted before integrating).
    """

    # sort the arrays and make sure they're numpy arrays
    [xdata, ydata] = sort_matrix([xdata,ydata],0)
    xdata = _n.array(xdata)
    ydata = _n.array(ydata)

    if xmin is None: xmin = min(xdata)
    if xmax is None: xmax = max(xdata)

    # find the index range
    imin = xdata.searchsorted(xmin)
    imax = xdata.searchsorted(xmax)

    xint = [xdata[imin]]
    yint = [0]

    # get the autozero
    if autozero >= 1:
        zero = _n.average(ydata[imin:imin+int(autozero)])
        ydata = ydata-zero

    for n in range(imin+1,imax):
        if len(yint):
            xint.append(xdata[n])
            yint.append(yint[-1]+0.5*(xdata[n]-xdata[n-1])*(ydata[n]+ydata[n-1]))
        else:
            xint.append(xdata[n])
            yint.append(0.5*(xdata[n]-xdata[n-1])*(ydata[n]+ydata[n-1]))

    return _n.array(xint), _n.array(yint)

def interpolate(xarray, yarray, x, rigid_limits=True):
    """

    returns the y value of the linear interpolated function
    y(x). Assumes increasing xarray!

    rigid_limits=False means when x is outside xarray's range,
    use the endpoint as the y-value.

    """
    if not len(xarray) == len(yarray):
        print("lengths don't match.", len(xarray), len(yarray))
        return None
    if x < xarray[0] or x > xarray[-1]:
        if rigid_limits:
            print("x=" + str(x) + " is not in " + str(min(xarray)) + " to " + str(max(xarray)))
            return None
        else:
            if x < xarray[0]: return yarray[0]
            else:             return yarray[-1]

    # find the index of the first value in xarray higher than x
    for n2 in range(1, len(xarray)):
        if x >= min(xarray[n2], xarray[n2-1]) and x <= max(xarray[n2], xarray[n2-1]):
            break
        if n2 == len(xarray):
            print("couldn't find x anywhere.")
            return None
    n1 = n2-1

    # now we have the indices surrounding the x value
    # interpolate!

    return yarray[n1] + (x-xarray[n1])*(yarray[n2]-yarray[n1])/(xarray[n2]-xarray[n1])



def invert_increasing_function(f, f0, xmin, xmax, tolerance, max_iterations=100):
    """
    This will try try to qickly find a point on the f(x) curve between xmin and xmax that is
    equal to f0 within tolerance.
    """

    for n in range(max_iterations):
        # start at the middle
        x = 0.5*(xmin+xmax)

        df = f(x)-f0
        if _n.fabs(df) < tolerance: return x

        # if we're high, set xmin to x etc...
        if df > 0: xmin=x
        else:      xmax=x

    print("Couldn't find value!")
    return 0.5*(xmin+xmax)


def is_close(x, array, fraction=0.0001):
    """

    compares x to all of the values in array.  If it's fraction close to
    any, returns true

    """

    result = False
    for n in range(0,len(array)):
        if array[n] == 0:
            if x == 0:
                result = True
        elif abs((x-array[n])/array[n]) < fraction:
            result = True

    return(result)





def join(array_of_strings, delimiter=' '):
    if array_of_strings == []: return ""

    if delimiter is None: delimiter=' '

    output = str(array_of_strings[0])
    for n in range(1, len(array_of_strings)):
        output += delimiter + str(array_of_strings[n])
    return(output)


def _load_object(path=None, text="Load a pickled object."):
    if path==None: path = _s.dialogs.SingleFile("*.pickle", text=text)
    if path == "": return None

    f = open(path, "r")
    object = _cPickle.load(f)
    f.close()

    object._path = path
    return object


def fft(t, y, pow2=False, window=None, rescale=False):
    """
    FFT of y, assuming complex or real-valued inputs. This goes through the 
    numpy fourier transform process, assembling and returning (frequencies, 
    complex fft) given time and signal data y.

    Parameters
    ----------
    t,y
        Time (t) and signal (y) arrays with which to perform the fft. Note the t
        array is assumed to be evenly spaced.
        
    pow2 = False
        Set this to true if you only want to keep the first 2^n data
        points (speeds up the FFT substantially)

    window = None
        Can be set to any of the windowing functions in numpy that require only
        the number of points as the argument, e.g. window='hanning'. Also accepts
        'None', False, 'False', 0, '0'
        
    rescale = False
        If True, the FFT will be rescaled by the square root of the ratio of 
        variances before and after windowing, such that the sum of component 
        amplitudes squared is equal to the actual variance.
    """
    # make sure they're numpy arrays, and make copies to avoid the referencing error
    y = _n.array(y)
    t = _n.array(t)

    # if we're doing the power of 2, do it
    if pow2:
        keep  = 2**int(_n.log2(len(y)))

        # now resize the data
        y.resize(keep)
        t.resize(keep)

    # Window the data
    if not window in [None, 'None', False, 'False', 0, '0']:
        try:
            # Get the windowing array
            w = eval("_n."+window, dict(_n=_n))(len(y))
            
            # Store the original variance
            v0 = _n.average(abs(y)**2)
            
            # window the time domain data 
            y = y * w
            
            # Rescale by the variance ratio
            if rescale: y = y * _n.sqrt(v0 / _n.average(abs(y)**2))
            
        except:
            print("ERROR: Bad window!")
            return

    # do the actual fft, and normalize
    Y = _n.fft.fftshift( _n.fft.fft(y) / len(t) )
    f = _n.fft.fftshift( _n.fft.fftfreq(len(t), t[1]-t[0]) )
    
    return f, Y
    





def psd(t, y, pow2=False, window=None, rescale=False):
    """
    Single-sided power spectral density, assuming real valued inputs. This goes 
    through the numpy fourier transform process, assembling and returning
    (frequencies, psd) given time and signal data y. 
    
    Note it is defined such that sum(psd)*df, where df is the frequency
    spacing, is the variance of the original signal for any range of frequencies.
    This includes the DC and Nyquist components: 
        
        sqrt(psd[0]*df)  = average value of original time trace
        
        sqrt(psd[-1]*df) = amplitude of Nyquist component (for even # points)
    
    Parameters
    ----------
    t,y
        Time (t) and signal (y) arrays with which to perform the PSD. Note the t
        array is assumed to be evenly spaced.

    pow2 = False
        Set this to true if you only want to keep the first 2^n data
        points (speeds up the FFT substantially)

    window = None
        can be set to any of the windowing functions in numpy,
        e.g. window='hanning'. Also accepts
        'None', False, 'False', 0, '0'
        
    rescale = False
        If True, the FFT will be rescaled by the square root of the ratio of 
        variances before and after windowing, such that the integral 
        sum(PSD)*df is the variance of the *original* time-domain data.

    returns frequencies, psd (y^2/Hz)
    """
    # do the actual fft
    f, Y = fft(t,y,pow2,window,rescale)
    
    # take twice the negative frequency branch, because it contains the 
    # extra frequency point when the number of points is odd.
    f = _n.abs(f[int(len(f)/2)::-1])
    P = _n.abs(Y[int(len(Y)/2)::-1])**2 / (f[1]-f[0])

    # Since this is the same as the positive frequency branch, double the
    # appropriate frequencies. For even number of points, there is one
    # extra negative frequency to avoid doubling. For odd, you only need to
    # avoid the DC value.
    
    # For the even
    if len(t)%2 == 0: P[1:len(P)-1] = P[1:len(P)-1]*2
    else:             P[1:]         = P[1:]*2

    return f, P


#if __name__ == '__main__':
#    t  = _n.linspace(0,10,1000)
#    y  = _n.cos(t*10)
#    yh = y*_n.hanning(len(y))
#    f, P = psd(t,y,True,'hanning',True)    


def read_lines(path):
    f = open(path, 'rU')
    a = f.readlines()
    f.close()

    return(a)

def replace_in_files(search, replace, depth=0, paths=None, confirm=True):
    """
    Does a line-by-line search and replace, but only up to the "depth" line.
    """

    # have the user select some files
    if paths==None:
        paths = _s.dialogs.MultipleFiles('DIS AND DAT|*.*')
    if paths == []: return

    for path in paths:
        lines = read_lines(path)

        if depth: N=min(len(lines),depth)
        else:     N=len(lines)

        for n in range(0,N):
            if lines[n].find(search) >= 0:
                lines[n] = lines[n].replace(search,replace)
                print(path.split(_os.path.pathsep)[-1]+ ': "'+lines[n]+'"')

        # only write if we're not confirming
        if not confirm:
            _os.rename(path, path+".backup")
            write_to_file(path, join(lines, ''))

    if confirm:
        if input("yes? ")=="yes":
            replace_in_files(search,replace,depth,paths,False)

    return

def replace_lines_in_files(search_string, replacement_line):
    """
    Finds lines containing the search string and replaces the whole line with
    the specified replacement string.
    """


    # have the user select some files
    paths = _s.dialogs.MultipleFiles('DIS AND DAT|*.*')
    if paths == []: return

    for path in paths:
        _shutil.copy(path, path+".backup")
        lines = read_lines(path)
        for n in range(0,len(lines)):
            if lines[n].find(search_string) >= 0:
                print(lines[n])
                lines[n] = replacement_line.strip() + "\n"
        write_to_file(path, join(lines, ''))

    return

def reverse(array):
    """
    returns a reversed numpy array
    """
    l = list(array)
    l.reverse()
    return _n.array(l)

def round_sigfigs(x, n=2):
    """
    Rounds the number to the specified significant figures. x can also be 
    a list or array of numbers (in these cases, a numpy array is returned).
    """
    iterable = is_iterable(x)
    if not iterable: x = [x]

    # make a copy to be safe
    x = _n.array(x)
    
    # loop over the elements
    for i in range(len(x)):
        
        # Handle the weird cases
        if not x[i] in [None, _n.inf, _n.nan]:
            sig_figs = -int(_n.floor(_n.log10(abs(x[i]))))+n-1
            x[i] = _n.round(x[i], sig_figs)
        
    if iterable: return x
    else:        return x[0]

def _save_object(object, path=None, text="Save this object where?"):
    if path==None: path = _s.dialogs.Save("*.pickle", text=text)
    if path == "": return

    if len(path.split(".")) <= 1 or not path.split(".")[-1] == "pickle":
        path = path + ".pickle"

    object._path = path

    f = open(path, "w")
    _cPickle.dump(object, f)
    f.close()

def shift_feature_to_x0(xdata, ydata, x0=0, feature=imax):
    """
    Finds a feature in the the ydata and shifts xdata so the feature is centered
    at x0. Returns shifted xdata, ydata. Try me with plot.tweaks.manipulate_shown_data()!

    xdata,ydata     data set
    x0=0            where to shift the peak
    feature=imax    function taking an array/list and returning the index of said feature
    """
    i = feature(ydata)
    return xdata-xdata[i]+x0, ydata

def smooth(array, index, amount):
    """

    Returns the average of the data at the specified index +/- amount

    """

    sum = array[index]
    count = 1.

    for n in range(1, amount+1):
        if index+n >= len(array):
            break
        sum   += array[index+n]
        count += 1.

    for n in range(1, amount+1):
        if index-n < 0:
            break
        sum   += array[index-n]
        count += 1.

    return(sum/count)

def smooth_array(array, amount=1):
    """

    Returns the nearest-neighbor (+/- amount) smoothed array.
    This does not modify the array or slice off the funny end points.

    """
    if amount==0: return array

    # we have to store the old values in a temp array to keep the
    # smoothing from affecting the smoothing
    new_array = _n.array(array)

    for n in range(len(array)):
        new_array[n] = smooth(array, n, amount)

    return new_array

def smooth_data(xdata, ydata, yerror, amount=1):
    """
    Returns smoothed [xdata, ydata, yerror]. Does not destroy the input arrays.
    """

    new_xdata  = smooth_array(_n.array(xdata), amount)
    new_ydata  = smooth_array(_n.array(ydata), amount)
    if yerror is None:  new_yerror = None
    else:               new_yerror = smooth_array(_n.array(yerror), amount)

    return [new_xdata, new_ydata, new_yerror]

def sort_matrix(a,n=0):
    """
    This will rearrange the array a[n] from lowest to highest, and
    rearrange the rest of a[i]'s in the same way. It is dumb and slow.

    Returns a numpy array.
    """
    a = _n.array(a)
    return a[:,a[n,:].argsort()] # this is magic.

def submatrix(matrix,i1,i2,j1,j2):
    """
    returns the submatrix defined by the index bounds i1-i2 and j1-j2

    Endpoints included!
    """

    new = []
    for i in range(i1,i2+1):
        new.append(matrix[i][j1:j2+1])
    return _n.array(new)

def trim_data(xmin, xmax, xdata, *args):
    """
    Removes all the data except that in which xdata is between xmin and xmax. 
    This does not mutilate the input arrays, and additional arrays
    can be supplied via args (provided they match xdata in shape)

    xmin and xmax can be None
    """

    # make sure it's a numpy array
    if not isinstance(xdata, _n.ndarray): xdata = _n.array(xdata)

    # make sure xmin and xmax are numbers
    if xmin is None: xmin = min(xdata)
    if xmax is None: xmax = max(xdata)

    # get all the indices satisfying the trim condition
    ns = _n.argwhere((xdata >= xmin) & (xdata <= xmax)).transpose()[0]

    # trim the xdata
    output = []
    output.append(xdata[ns])

    # trim the rest
    for a in args:
        # make sure it's a numpy array
        if not isinstance(a, _n.ndarray): a = _n.array(a)
        output.append(a[ns])

    return output

def trim_data_uber(arrays, conditions):
    """
    Non-destructively selects data from the supplied list of arrays based on 
    the supplied list of conditions. Importantly, if any of the conditions are 
    not met for the n'th data point, the n'th data point is rejected for
    all supplied arrays.

    Example
    -------
    x = numpy.linspace(0,10,20)
    y = numpy.sin(x)
    trim_data_uber([x,y], [x>3,x<9,y<0.7])
      
    This will keep only the x-y pairs in which 3<x<9 and y<0.7, returning
    a list of shorter arrays (all having the same length, of course).
    """    
    # dumb conditions
    if len(conditions) == 0: return arrays
    if len(arrays)     == 0: return []    
    
    # find the indices to keep
    all_conditions = conditions[0]
    for n in range(1,len(conditions)): all_conditions = all_conditions & conditions[n]
    ns = _n.argwhere(all_conditions).transpose()[0]

    # assemble and return trimmed data    
    output = []
    for n in range(len(arrays)): 
        if not arrays[n] is None: output.append(arrays[n][ns])
        else:                     output.append(None)
    return output
    


def ubersplit(s, delimiters=['\t','\r',' ']):

    # run through the string, replacing all the delimiters with the first delimiter
    for d in delimiters: s = s.replace(d, delimiters[0])
    return s.split(delimiters[0])

def write_to_file(path, string):
    file = open(path, 'w')
    file.write(string)
    file.close()

def _match_data_sets(x,y):
    """
    Makes sure everything is the same shape. "Intelligently".
    
    Parameters
    ----------
    x,y
        Can be a single array, list of numbers, a number, None, or lists of 
        any combination of these.
    
    Returns
    -------
    matched_x, matched_y
        Each is a list of arrays of numbers, with all dimensions matched.
    """

    # If x is a value, use this for the step size.
    dx = 1.0
    if is_a_number(x):
        dx = x
        x  = None

    # Handle the None for x or y
    if x is None or len(x) == 0:
        # If x is None, y can be either [1,2] or [[1,2],[1,2]] or []
        if len(y) == 0: x = []

        elif is_iterable(y[0]):
            # make an array of arrays to match
            x = []
            for n in range(len(y)):
                x.append(list(dx*_n.array(range(len(y[n])))))
        else: x = list(range(len(y)))

    # If y is a value, use this for the step size.
    dy = 1.0
    if is_a_number(y):
        dy = y
        y  = None

    if y is None or len(y) == 0:
        # If y is none, x can be either [1,2] or [[1,2],[1,2]] or []
        if len(x) == 0: y = []

        elif is_iterable(x[0]):
            # make an array of arrays to match
            y = []
            for n in range(len(x)):
                y.append(list(dy*_n.array(range(len(x[n])))))
        else: y = list(range(len(x)))

    # At this point they should be matched, but may still be 1D
    # Default behavior: if all elements are numbers in both, assume they match
    if elements_are_numbers(x): x=[x]
    if elements_are_numbers(y): y=[y]

    # Make sure they are both lists (so append works!)
    if not type(x) == list: x = list(x)
    if not type(y) == list: y = list(y)

    # Make sure they're the same length
    while len(x) > len(y): y.append(y[-1])
    while len(y) > len(x): x.append(x[-1])

    # Second default behavior: shared array [1,2,3], [[1,2,1],[1,2,1]] or vis versa
    if elements_are_numbers(x) and not elements_are_numbers(y): x = [x]*len(y)
    if elements_are_numbers(y) and not elements_are_numbers(x): y = [y]*len(x)

    # Clean up any remaining Nones
    for n in range(len(x)):
        if x[n] is None: x[n] = list(range(len(y[n])))
        if y[n] is None: y[n] = list(range(len(x[n])))

    return x, y

def _match_error_to_data_set(x, ex):
    """
    Inflates ex to match the dimensionality of x, "intelligently".
    x is assumed to be a 2D array.
    
    Returns
    -------
    The inflated / matched version of ex.
    """
    # Simplest case, ex is None or a number
    if not is_iterable(ex):

        # Just make a matched list of Nones
        if ex is None: ex = [ex]*len(x)

        # Make arrays of numbers
        if is_a_number(ex):
            value = ex # temporary storage
            ex    = []
            for n in range(len(x)):
                ex.append([value]*len(x[n]))

    # At this point, ex is iterable

    # Default behavior: If the elements are all numbers and the length matches
    # that of the first x-array, assume this is meant to match all the x
    # data sets
    if elements_are_numbers(ex) and len(ex) == len(x[0]): ex = [ex]*len(x)

    # Make sure it's a list (for appending)
    if not type(ex) == list: ex = list(ex)

    # The user may specify a list of some iterable and some not.
    # The list length may not match
    while len(ex) < len(x): ex.append(ex[-1])

    # Now they are the same length
    for n in range(len(x)):
        # do nothing to the None's
        # Inflate single numbers to match
        if is_a_number(ex[n]): ex[n] = [ex[n]]*len(x[n])

    return ex

    
if __name__ == '__main__':  
    import spinmob
    runfile(spinmob.__path__[0] + '/tests/test__functions.py')
     